////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Byte stream implementation
///
///     This library abstracts a user buffer into a byte stream.  At the
///     expense of performance, this library helps clean up user interfaces
///     that have buffer requirements (e.g. user doesn't have to pass
///     around buf/buf_len pairs, but instead pass a byte stream struct).
///     Note that this library assumes that the user is responsible for
///     endian issues.
///
///     Note the byte stream offset tracks the position within the stream
///     to read/write next byte.  The offset is constrained to be
///     0<=offset<=size, where size is the number of bytes in the stream.
///     Once the stream is completely utilized (either full of written data
///     or consumed), offset=size (offset indexes one byte past the stream
///     window).  This is used throughout the functions as a consistency
///     check and in tera_bstream_get_num_left.
//
////////////////////////////////////////////////////////////////////////////


#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_bstream.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  tera_bstream_init
///     Initialize the byte stream
//
//  Description:
///     This function initializes the byte stream.  The user must provide
///     a buffer which is abstracted by the byte stream.
///
/// \return
///     None
///
void tera_bstream_init(
    sTERA_BSTREAM_CBLK  *cb,            ///< pointer to bstream control block
    UINT8               *buf,           ///< pointer to user buffer
    UINT32               buf_size       ///< size of user buffer (in bytes)
    )
{
    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(buf != NULL);
    ASSERT(buf_size != 0);

    // store the pointer to user buffer
    cb->buf = buf;

    // store the user buffer size
    cb->buf_size = buf_size;

    // reset the byte stream offset
    cb->offset = 0;
}

// --------------------------------------------------------------------------
//  tera_bstream_set_curpos
///     Set current position in byte stream
//
//  Description:
///     This function sets the current byte stream offset.
///
/// \return
///     None
///
void tera_bstream_set_curpos(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to bstream control block
    UINT32                  cur_pos     ///< current position in byte stream
    )
{
    // consistency checks
    ASSERT(cb != NULL);

    // consistency check that current position is within byte stream window
    ASSERT(cur_pos <= cb->buf_size);

    // set the offset
    cb->offset = cur_pos;
}

// --------------------------------------------------------------------------
//  tera_bstream_get_curpos
///     Get current position in byte stream
//
//  Description:
///     This function gets the current byte stream offset.
///
/// \return
///     Current offset position
///
UINT32 tera_bstream_get_curpos(
    sTERA_BSTREAM_CBLK  *cb             ///< pointer to bstream control block
    )
{
    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(cb->offset <= cb->buf_size);

    return(cb->offset);
}

// --------------------------------------------------------------------------
//  tera_bstream_get_num_left
///     Get number of byte stream bytes left to be processed
//
//  Description:
///     This function returns the number of bytes left in the byte stream.
///
/// \return
///     Number of bytes left in byte stream
///
UINT32 tera_bstream_get_num_left(
    sTERA_BSTREAM_CBLK *cb              ///< pointer to bstream control block
    )
{
    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(cb->offset <= cb->buf_size);

    return(cb->buf_size - cb->offset);
}

// --------------------------------------------------------------------------
//  tera_bstream_getraw
///     Get multi-byte raw data from byte stream
//
//  Description:
///     This function obtains an array of bytes from the byte stream
///     and advances the byte stream offset.  The user must provide
///     destination memory large enough to hold the data.
///
/// \return
///     None
///
void tera_bstream_getraw(
    sTERA_BSTREAM_CBLK *cb,             ///< pointer to bstream control block
    UINT32              num_bytes,      ///< number of bytes to store in dst
    UINT8              *dst             ///< pointer to destination memory
    )
{
    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(num_bytes != 0);
    ASSERT(dst != NULL);

    // consistency check that getting data from the byte stream will not
    // go past the end of the byte stream
    ASSERT((cb->offset+num_bytes) <= cb->buf_size);

    // get raw data from byte stream, store in user provided destination
    memcpy(dst,
           &cb->buf[cb->offset],
           num_bytes);

    // update byte stream offset with the number of bytes we've read
    (cb->offset)+= num_bytes;
}

// --------------------------------------------------------------------------
//  tera_bstream_get8
///     Get 8-bit value from byte stream
//
//  Description:
///     This function reads a 8-bit value from the byte stream and advances
///     the byte stream offset.
///
/// \return
///     8-bit value
///
UINT8 tera_bstream_get8(
    sTERA_BSTREAM_CBLK  *cb             ///< pointer to bstream control block
    )
{
    UINT8   val;

    // read the 8-bit value from byte stream
    tera_bstream_getraw(cb,
                        sizeof(UINT8),
                        &val);

    return val;
}

// --------------------------------------------------------------------------
//  tera_bstream_get16
///     Get 16-bit value from byte stream
//
//  Description:
///     This function reads a 16-bit value from the byte stream and advances
///     the byte stream offset.
///
/// \return
///     16-bit value
///
UINT16 tera_bstream_get16(
    sTERA_BSTREAM_CBLK *cb              ///< pointer to bstream control block
    )
{
    UINT16   val;

    // read the 16-bit value from byte stream
    tera_bstream_getraw(cb,
                        sizeof(UINT16),
                        (UINT8 *)&val);

    // return value
    return(val);
}

// --------------------------------------------------------------------------
//  tera_bstream_get32
///     Get 32-bit value from byte stream
//
//  Description:
///     This function reads a 32-bit value from the byte stream and advances
///     the byte stream offset.
///
/// \return
///     32-bit value
///
UINT32 tera_bstream_get32(
    sTERA_BSTREAM_CBLK *cb              ///< pointer to bstream control block
    )
{
    UINT32   val;

    // read the 32-bit value from byte stream
    tera_bstream_getraw(cb,
                        sizeof(UINT32),
                        (UINT8 *)&val);

    // return value
    return(val);
}

// --------------------------------------------------------------------------
//  tera_bstream_putraw
///     Put multi-byte raw data from source into byte stream
//
//  Description:
///     This function stores an array of bytes into the byte stream
///     and advances the byte stream offset.
///
/// \return
///     None
///
void tera_bstream_putraw(
    sTERA_BSTREAM_CBLK  *cb,            ///< pointer to bstream control block
    UINT32               num_bytes,     ///< number of bytes to store
    UINT8               *data           ///< pointer to source data
    )
{
    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(num_bytes != 0);
    ASSERT(data != NULL);

    // consistency check that adding data to the byte stream will not
    // go past the end of the byte stream
    ASSERT((cb->offset+num_bytes) <= cb->buf_size);

    // store raw data in byte stream
    memcpy(&cb->buf[cb->offset],
           data,
           num_bytes);

    // update byte stream offset with the number of bytes we've written
    (cb->offset)+= num_bytes;
}

// --------------------------------------------------------------------------
//  tera_bstream_put8
///     Put 8-bit value in byte stream
//
//  Description:
///     This function stores a 8-bit value in the byte stream and advances
///     the byte stream offset.
///
/// \return
///     None
///
void tera_bstream_put8(
    sTERA_BSTREAM_CBLK *cb,             ///< pointer to bstream control block
    UINT8               val             ///< value to store
    )
{
    // write the 8-bit value in byte stream
    tera_bstream_putraw(cb,
                        sizeof(UINT8),
                        &val);
}

// --------------------------------------------------------------------------
//  tera_bstream_put16
///     Put 16-bit value in byte stream
//
//  Description:
///     This function stores a 16-bit value in the byte stream and advances
///     the byte stream offset.
///
/// \return
///     None
///
void tera_bstream_put16(
    sTERA_BSTREAM_CBLK *cb,             ///< pointer to bstream control block
    UINT16              val             ///< value to store
    )
{
    // write the 16-bit value in byte stream
    tera_bstream_putraw(cb,
                        sizeof(UINT16),
                        (UINT8 *)&val);
}

// --------------------------------------------------------------------------
//  tera_bstream_put32
///     Put 32-bit value in byte stream
//
//  Description:
///     This function stores a 32-bit value in the byte stream and advances
///     the byte stream offset.
///
/// \return
///     None
///
void tera_bstream_put32(
    sTERA_BSTREAM_CBLK *cb,             ///< pointer to bstream control block
    UINT32              val             ///< value to store
    )
{
    // write the 32-bit value in byte stream
    tera_bstream_putraw(cb,
                        sizeof(UINT32),
                        (UINT8 *)&val);
}

