////////////////////////////////////////////////////////////////////////////
//   Copyright 2010  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Memory Module
///
///     This file defines the memory module API prototypes and defintions.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include <linux/pci.h>

// ---- Constants and Types -------------------------------------------------

// -------------------------------------------------------------------------- 
//  sTERA_MEM_BUF
///     Memory buffer
//
//  Description:
///     This structure defines a memory buffer allocated from the memory pool.
///     The memory buffer points to DMA'able memory, so it has both a virtual
///     address for host OS software, and a physical address for hardware
///     accesses.
///
typedef struct sTERA_MEM_BUF_STRUCT
{
    void            *virt_addr;    ///< Virtual address of buffer
    dma_addr_t       phys_addr;    ///< Physical address of buffer
    INT              avail;        ///< flag indicates whether buf is available
} sTERA_MEM_BUF;

// -------------------------------------------------------------------------- 
//  sTERA_MEM_DMA
///     DMA'able memory
//
//  Description:
///     This structure defines an allocation of DMA'able memory. This memory
///     does NOT come from the memory pool. The memory can be accessed by
///     software via the virtual address, or by hardware via the physical
///     address.
///
typedef struct sTERA_MEM_HANDLE_STRUCT
{
    void            *virt_addr;    ///< Virtual address of memory
    dma_addr_t       phys_addr;    ///< Physical address of memory
    size_t           len;          ///< Total length of memory
    void            *sys;          ///< System opaque pointer
    void            *cblk;         ///< Memory module control block
} sTERA_MEM_DMA;

// Defines for memory pool size
#define TERA_CHUNK_SIZE    PAGE_SIZE
#define TERA_NUM_CHUNKS    256
#define TERA_MEM_POOL_SIZE (TERA_NUM_CHUNKS * TERA_CHUNK_SIZE)

// -------------------------------------------------------------------------- 
//  sTERA_POOL_CBLK
///     Memory pool control block
//
//  Description:
///     This structure defines the memory pool module control block. The
///     control block contains the state flag and the memory pool.
///
#if 0
typedef struct sTERA_POOL_CBLK_STRUCT
{
    BOOLEAN        initialized;  ///< Initialized flag
    sTERA_MEM_BUF  chunk[TERA_NUM_CHUNKS];
} sTERA_POOL_CBLK;
#endif

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

extern INT32 tera_mem_init(
    struct pci_dev *dev
    );

extern INT32 tera_mem_exit(
    struct pci_dev *dev
    );

extern sTERA_MEM_DMA *tera_mem_alloc(
    struct pci_dev *dev,
    size_t          len
    );

extern void tera_mem_free(
    struct pci_dev *dev,
    sTERA_MEM_DMA  *buf
    );

extern sTERA_MEM_BUF  *tera_mem_buf_alloc(
    struct pci_dev *dev,
    size_t          len
    );

extern void tera_mem_buf_free(
    struct pci_dev *dev,
    sTERA_MEM_BUF  *buf
    );
