////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Packet Module Implementation
///
///     This file implements the packet module, which implements the packet
///     queues, queue entries, and operations on the queues.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_pkt.h"
#include "tera_pkt_internal.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// Message pool name string table
static const char *pkt_pool_name_str_tbl[MAX_NUM_GFUN] = {
    "pcoipHostPktPool0",
    "pcoipHostPktPool1"
};

// ---- Local Function Prototypes -------------------------------------------

static sTERA_PKT *pkt_alloc(
    struct pci_dev *dev,
    sPKT_CBLK      *cblk,
    UINT32          buf_len
    );

// --------------------------------------------------------------------------
//  pkt_alloc
//      Allocate one packet
//
//  Description:
//      This function allocate one packet with the specific buffer length. The
//      memory for the packet structure is allocated from a lookaside list, and
//      the buffer is allocated from the memory pool. If there is not enough
//      memory the the function returns NULL.
//
static sTERA_PKT *pkt_alloc(
    struct pci_dev *dev,    // Device object
    sPKT_CBLK      *cblk,   // Control block
    UINT32          buf_len // Buffer length
    )
{
    sTERA_PKT *ret = NULL;

    ASSERT(cblk != NULL);

    // Create a new buffer from lookaside list for the packet structure
    ret = kmem_cache_alloc(cblk->pkt_mem_pool, GFP_KERNEL);
    if (!ret)
    {
        mTERA_LOG_ERR("kmem_cache_alloc failed");
    }
    else
    {
        // Initialize the packet structure fields
        memset(ret, 0, sizeof(sTERA_PKT));
        ret->buf_len = buf_len;

        // Allocate some DMA'able memory for the packet buffer
        ret->buf = tera_mem_buf_alloc(dev, ret->buf_len);
        if (ret->buf == NULL)
        {
            mTERA_LOG_ERR("failed to allocate packet buffer");

            kmem_cache_free(cblk->pkt_mem_pool, ret);
            ret = NULL;
        }
        else
        {
            // Increment the packet allocate counter
            cblk->pkt_alloc_cnt++;
        }
    }

    return ret;
} // pkt_alloc()

// --------------------------------------------------------------------------
//  tera_pkt_init
///     Initialize the packet module
//
//  Description:
///     This function initializes the packet module.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_pkt_init(
    struct pci_dev *dev,    // Device object
    UINT32          dev_id  // Device ID
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sPKT_CBLK      *cblk = &dev_cblk->pkt_cblk;
    INT32           ret;

    // Create a lookaside list for buffers for packets
    cblk->pkt_mem_pool = kmem_cache_create(pkt_pool_name_str_tbl[dev_id],
                                           sizeof(sTERA_PKT),
                                           0, SLAB_HWCACHE_ALIGN,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
                                           NULL);
#else
                                           NULL, NULL);
#endif
    if (!cblk->pkt_mem_pool)
    {
        cblk->initialized = 0;
        ret = TERA_ERR_NO_MEMORY;
        mTERA_LOG_CRIT("kmem_cache_create failed");
    }
    else
    {
        // Initialize the control block fields
        cblk->initialized = 1;
        cblk->pkt_alloc_cnt = 0;
        cblk->pkt_free_cnt = 0;
        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_pkt_init()

// --------------------------------------------------------------------------
//  tera_pkt_cleanup
///     Clean up the packet module
//
//  Description:
///     This function cleans up the packet module. This function marks the
///     module as uninitialized. All operations on packets and packet queues,
///     such as freeing or purging, should be done before calling this
///     function.
///
/// \return
///     None
///
void tera_pkt_cleanup(
    struct pci_dev *dev     // Device object
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);

    mTERA_LOG_DBG("start");

    if (dev_cblk->pkt_cblk.initialized)
    {
        mTERA_LOG_INFO("pkt_alloc_cnt=%u, pkt_free_cnt=%u",
                       dev_cblk->pkt_cblk.pkt_alloc_cnt,
                       dev_cblk->pkt_cblk.pkt_free_cnt);

        // Mark the module as uninitialized
        dev_cblk->pkt_cblk.initialized = 0;

        // destroy the lookaside cache
        kmem_cache_destroy(dev_cblk->pkt_cblk.pkt_mem_pool);
    }
    else
    {
        mTERA_LOG_ERR("Unexpected state or invalid parameter");
    }

    mTERA_LOG_DBG("end");
} // tera_pkt_cleanup()

// --------------------------------------------------------------------------
//  tera_pkt_alloc
///     Allocate one or more packets
//
//  Description:
///     This function allocates one or more packets with the specific buffer
///     length. If there is not enough memory for the requested number of
///     packets, then no packets will be returned. If successful, the requested
///     number of packets are returned, chained one after another via the next
///     pointer. The packet buffers are zeroed out.
///
/// \return
///     Pointer to a packet,
///     NULL
///
sTERA_PKT *tera_pkt_alloc(
    struct pci_dev *dev,        ///< Device object
    UINT32          buf_len,    ///< Buffer length
    UINT32          num_pkt     ///< Number of packets to allocate
    )
{
    sTERA_PKT      *ret = NULL;
    sTERA_PKT     **pkt;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sPKT_CBLK      *cblk = &dev_cblk->pkt_cblk;
    UINT32          i;

    if (cblk->initialized)
    {
        pkt = &ret;

        for (i = 0; i < num_pkt; i++)
        {
            *pkt = pkt_alloc(dev, cblk, buf_len);

            if (*pkt == NULL)
            {
                mTERA_LOG_ERR("failed to allocate packet %u of %u", i + 1,
                              num_pkt);

                // Free all the buffers
                tera_pkt_free(dev, ret);

                ret = NULL;
                break;
            }

            pkt = &(*pkt)->next;
        }
    } // end if

    return ret;
} // tera_pkt_alloc()

// --------------------------------------------------------------------------
//  tera_pkt_free
///     Free the memory used by the packet(s)
//
//  Description:
///     This function frees the memory that was used by the packet(s). If the
///     packet buffer is not NULL, then the buffer is freed as well. It is
///     important that the buffer is not used elsewhere after calling this
///     function. The caller can pass in multiple packets by chaining the
///     packets via the next pointer.
///
/// \return
///     None
///
void tera_pkt_free(
    struct pci_dev *dev,    ///< Device object
    sTERA_PKT *pkt          ///< Packet
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sPKT_CBLK      *cblk = &dev_cblk->pkt_cblk;
    sTERA_PKT      *curr;
    sTERA_PKT      *next;

    if (cblk->initialized && (pkt != NULL))
    {
        curr = pkt;

        while (curr != NULL)
        {
            next = curr->next;

            if (curr->buf != NULL)
            {
                // Delete the DMA'able packet buffer
                tera_mem_buf_free(dev, curr->buf);
            }

            // Delete the memory object of which the buffer belongs to
            kmem_cache_free(cblk->pkt_mem_pool, curr);

            // Increment the packet free counter
            cblk->pkt_free_cnt++;

            // Move on to the next packet
            curr = next;
        }
    }
} // tera_pkt_free()

// --------------------------------------------------------------------------
//  tera_pkt_q_init
///     Initialize a packet queue
//
//  Description:
///     This function initializes a packet queue by creating the spinlock
///     that is used for accessing the queue, and also initializing the head
///     pointer. A packet queue MUST be initialized before performing enqueuing
///     or dequeuing packets to/from it.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_pkt_q_init(
    sTERA_PKT_Q *pkt_q    ///< Packet queue
    )
{
    INT32   ret = TERA_ERR_INVALID_ARG;

    if (pkt_q != NULL)
    {
        // Initialize the message queue variables
        init_MUTEX(&pkt_q->lock);
        pkt_q->head = NULL;
        pkt_q->initialized = 1;
        pkt_q->enqueue_cnt = 0;
        pkt_q->dequeue_cnt = 0;

        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_pkt_q_init()

// --------------------------------------------------------------------------
//  tera_pkt_q_purge
///     Purge all packets from the packet queue
//
//  Description:
///     This function purges all packets from the packet queue. If any packets
///     in the queue contain a DMA'able memory buffer, that buffer is freed as
///     well. This function asserts if the packet queue is NULL or not
///     initialized.
///
void tera_pkt_q_purge(
    struct pci_dev *dev,    ///< Device object
    sTERA_PKT_Q    *pkt_q   ///< Packet queue
    )
{
    sTERA_PKT *pkt;
    UINT32     purge_cnt = 0;

    ASSERT(pkt_q != NULL);
    ASSERT(pkt_q->initialized);

    while ((pkt = tera_pkt_q_dequeue(pkt_q)) != NULL)
    {
        // Free the packet
        tera_pkt_free(dev, pkt);

        purge_cnt++;
    }

    mTERA_LOG_DBG("enqueue_cnt=%u, dequeue_cnt=%u",
                  pkt_q->enqueue_cnt, pkt_q->dequeue_cnt);

    mTERA_LOG_DBG("purged %u packets", purge_cnt);
} // tera_pkt_q_purge()

// --------------------------------------------------------------------------
//  tera_pkt_q_enqueue
///     Enqueue one or more packets to the tail of the packet queue
//
//  Description:
///     This function enqueues one or more packets to the tail of the specified
///     packet queue. The user can enqueue more than one packet by chaining
///     the packets using the next pointer. The last packet in the chain MUST
///     have a NULL next pointer. This function is synchronized to other access
///     attempts to the packet queue. This function asserts if the packet queue
///     is NULL or not initialized.
///
/// \return
///     None
///
void tera_pkt_q_enqueue(
    sTERA_PKT_Q *pkt_q,    ///< Packet queue
    sTERA_PKT   *pkt       ///< Packet
    )
{
    sTERA_PKT *node;

    ASSERT(pkt_q != NULL);
    ASSERT(pkt_q->initialized);

    if (pkt != NULL)
    {
#ifdef DBG
        // Count how many packets we are going to enqueue
        node = pkt;

        do
        {
            pkt_q->enqueue_cnt++;

            node = node->next;
        } while (node != NULL);
#endif

        down(&pkt_q->lock);

        if (pkt_q->head == NULL)
        {
//            mTERA_LOG_DBG("enqueuing to head of packet queue");

            pkt_q->head = pkt;
        }
        else
        {
            node = pkt_q->head;

//            mTERA_LOG_DBG("enqueuing to tail of packet queue");

            while (node->next != NULL)
            {
                node = node->next;
            }

            node->next = pkt;
        }

//        mTERA_LOG_DBG("enqueue_cnt=%u", pkt_q->enqueue_cnt);

        up(&pkt_q->lock);
    } // end if
} // tera_pkt_q_enqueue()

// --------------------------------------------------------------------------
//  tera_pkt_q_head_enqueue
///     Enqueue one or more packets to the head of the packet queue
//
//  Description:
///     This function enqueues one or more packets to the head of the specified
///     packet queue. The user can enqueue more than one packet by chaining the
///     packets using the next pointer. The last packet in the chain MUST have
///     a NULL next pointer. This function is synchronized to other access
///     attempts to the packet queue. This function asserts if the packet queue
///     is NULL or not initialized.
///
/// \return
///     None
///
void tera_pkt_q_head_enqueue(
    sTERA_PKT_Q *pkt_q,    ///< Packet queue
    sTERA_PKT   *pkt       ///< Packet
    )
{
    sTERA_PKT *last = pkt;

    ASSERT(pkt_q != NULL);
    ASSERT(pkt_q->initialized);

    if (pkt != NULL)
    {
        for (;;)
        {
#ifdef DBG
            pkt_q->enqueue_cnt++;
#endif

            if (last->next == NULL)
            {
                // We've reached the last packet
                break;
            }

            last = last->next;
        }

        down(&pkt_q->lock);

        if (pkt_q->head == NULL)
        {
            pkt_q->head = pkt;
        }
        else
        {
            //
            // Set the last packet in the chain to point to the current packet
            // at the head, and then point the head to the start of the packet
            // chain
            //
            last->next = pkt_q->head;
            pkt_q->head = pkt;
        }

//        mTERA_LOG_DBG("enqueue_cnt=%u", pkt_q->enqueue_cnt);

        up(&pkt_q->lock);
    } // end if
} // tera_pkt_q_head_enqueue()

// --------------------------------------------------------------------------
//  tera_pkt_q_dequeue
///     Dequeue a packet from the head of the specified packet queue
//
//  Description:
///     This function dequeues a packet from the head of the specified packet
///     queue. If the packet queue is empty, this function returns NULL.
///
///     This function may be called after this module has been initialized. It
///     does not care if the module has been started or stopped.
///
/// \return
///     Pointer to a packet,
///     NULL
///
sTERA_PKT *tera_pkt_q_dequeue(
    sTERA_PKT_Q *pkt_q    ///< Packet queue
    )
{
    sTERA_PKT *ret = NULL;

    ASSERT(pkt_q != NULL);
    ASSERT(pkt_q->initialized);

    down(&pkt_q->lock);

    if (pkt_q->head != NULL)
    {
        //
        // Take the packet off of the head of the packet queue and detach it
        // from the queue by assigning its next pointer to NULL
        //
        ret = pkt_q->head;
        pkt_q->head = (pkt_q->head)->next;
        ret->next = NULL;

#ifdef DBG
        pkt_q->dequeue_cnt++;
#endif

//        mTERA_LOG_DBG("dequeue_cnt=%u, buf_len=%d", pkt_q->dequeue_cnt, ret->buf_len);
    }

    up(&pkt_q->lock);

    return ret;
} // tera_pkt_q_dequeue()

