////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Simple Communications Protocol Module Implementation
///
///     This file implements the SCP module.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_scp.h"
#include "tera_scp_internal.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

//
// Supported pri number (Only one value is supported because this driver does
// not support multiple users)
//
#define SUPPORTED_PRI_NUM  0

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

static INT32 scp_ok_pkt_send(
    struct pci_dev *dev,
    sSCP_INSTANCE  *instance,
    UINT8           chan_id,
    sTERA_PKT_Q    *pkt_q
    );

static INT32 scp_pkt_send(
    struct pci_dev *dev,
    UINT8           chan_id,
    UINT16          mtu,
    sTERA_PKT_Q    *pkt_q,
    eSCP_PKT_TYPE   pkt_type,
    UINT8          *apdu,
    UINT32          apdu_size
    );

// --------------------------------------------------------------------------
//  scp_ok_pkt_send
///     Send a SCP OK packet
//
//  Description:
///     This function sends a SCP OK packet. This packet is used during the
///     initial handshaking between the driver and the firmware when negotiating
///     the SCP channel. The packet is enqueue onto the packet queue.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NO_MEMORY
///
static INT32 scp_ok_pkt_send(
    struct pci_dev *dev,    // Device object
    sSCP_INSTANCE  *instance,
    UINT8           chan_id,
    sTERA_PKT_Q    *pkt_q
    )
{
    INT32 ret;
    UINT8 data;

    ASSERT(instance != NULL);
    ASSERT(chan_id < TERA_SCP_NUM_CHAN);
    ASSERT(pkt_q != NULL);

    // Fill out the data of the packet, which is the channel ID
    data = chan_id;

    // Send the SCP OK packet
    ret = scp_pkt_send(dev, chan_id, instance->mtu, pkt_q,
                       SCP_PKT_TYPE_OK, &data, sizeof(data));
    if (ret == TERA_SUCCESS)
    {
        // Mark the channel as ready
        instance->ready |= 1 << chan_id;
        mTERA_LOG_DBG("success sending SCP OK packet: chan=%d, ready=0x%X", chan_id, instance->ready);
    }
    else
    {
        mTERA_LOG_ERR("error sending SCP OK packet: ret=%d, chan=%d", ret, chan_id);
    }

    return ret;
} // scp_ok_pkt_send()

// --------------------------------------------------------------------------
//  scp_pkt_send
//      Send a SCP packet
//
//  Description:
//      This function sends a SCP packet. This function creates a packet, fills
//      out the SCP and SAR headers, and copies the user buffer into the packet
//      buffer. The completed packet is then put into the specified packet
//      queue.
//
static INT32 scp_pkt_send(
    struct pci_dev *dev,         // Device object
    UINT8           chan_id,     // SCP channel ID
    UINT16          mtu,         // MTU
    sTERA_PKT_Q    *pkt_q,       // Packet queue
    eSCP_PKT_TYPE   pkt_type,    // SCP packet type
    UINT8          *apdu,        // APDU
    UINT32          apdu_size    // APDU size
    )
{
    INT32         ret = TERA_SUCCESS;
    sTERA_PKT    *pkt;
    sTERA_PKT    *curr_pkt;
    sSCP_PKT_HDR *scp_pkt_hdr;
    sSAR_PKT_HDR *sar_pkt_hdr;
    UINT8        *data;
    UINT32        max_data_len;
    UINT32        num_pkt = 1;
    UINT32        bytes_rem = apdu_size;
    UINT32        num_bytes;
    UINT32        i;
    eAPDU_SEG_LOC apdu_seg_loc;

    ASSERT(chan_id < TERA_SCP_NUM_CHAN);
    ASSERT(pkt_q != NULL);
    ASSERT((apdu_size == 0) || (apdu != NULL));

    //
    // Calculate the maximum data size, which is the MTU minus the SCP and
    // SAR header
    //
    max_data_len = mtu - sizeof(sSCP_PKT_HDR) - sizeof(sSAR_PKT_HDR);

    // See if we need more than one packets
    if (apdu_size > 0)
    {
        num_pkt = (apdu_size + max_data_len - 1) / max_data_len;
    }

    //
    // Allocate all the packets that we need first - this is the only place in
    // this function that may cause us to return an error. After this point,
    // all operations are guaranteed to succeed.
    //
    pkt = tera_pkt_alloc(dev, mtu, num_pkt);
    if (pkt == NULL)
    {
        mTERA_LOG_ERR("failed to allocate packet(s)");

        ret = TERA_ERR_NO_MEMORY;
    }
    else
    {
//        mTERA_LOG_DBG("allocated %u packet(s)", num_pkt);

        // Fill out the packets
        for (i = 0, curr_pkt = pkt; i < num_pkt; i++)
        {
            // Figure out the APDU segment location
            apdu_seg_loc = (num_pkt == 1) ? APDU_SEG_LOC_ONLY_ONE :
                           (i == 0) ? APDU_SEG_LOC_FIRST :
                           (i == (num_pkt - 1)) ? APDU_SEG_LOC_LAST :
                           APDU_SEG_LOC_MID;

#if 1 //DBG
            switch (apdu_seg_loc)
            {
                case APDU_SEG_LOC_ONLY_ONE:
                {
                    mTERA_LOG_DBG("APDU_SEG_LOC_ONLY_ONE");
                    break;
                }
                case APDU_SEG_LOC_FIRST:
                {
                    mTERA_LOG_DBG("APDU_SEG_LOC_FIRST");
                    break;
                }
                case APDU_SEG_LOC_LAST:
                {
                    mTERA_LOG_DBG("APDU_SEG_LOC_LAST");
                    break;
                }
                case APDU_SEG_LOC_MID:
                {
                    mTERA_LOG_DBG("APDU_SEG_LOC_MID");
                    break;
                }
            }
#endif

            // Figure out how many bytes to copy
            num_bytes = (bytes_rem > max_data_len) ? max_data_len : bytes_rem;
            bytes_rem -= num_bytes;

            // Fill out the SCP packet header
            scp_pkt_hdr = mSCP_PKT_HDR_PTR_GET(curr_pkt->buf);
            scp_pkt_hdr->chan_id = chan_id;
            scp_pkt_hdr->pri_num = SUPPORTED_PRI_NUM;
            scp_pkt_hdr->pkt_type = pkt_type;

            // Fill out the SAR packet header
            sar_pkt_hdr = mSAR_PKT_HDR_PTR_GET(curr_pkt->buf);
            sar_pkt_hdr->apdu_size = mTERA_UINT32_BYTE_SWAP(apdu_size);
            sar_pkt_hdr->apdu_seg_size = mTERA_UINT16_BYTE_SWAP(num_bytes);
            sar_pkt_hdr->apdu_seg_loc = apdu_seg_loc;

#if 0
            mTERA_LOG_DBG("scp: chan_id=%u, pri_num=%u, pkt_type%u",
                          scp_pkt_hdr->chan_id, scp_pkt_hdr->pri_num,
                          scp_pkt_hdr->pkt_type);

            mTERA_LOG_DBG("sar: apdu_size=%u, apdu_seg_size=%u",
                          apdu_size, num_bytes);
#endif

            // Fill out the data
            data = mDATA_PTR_GET(curr_pkt->buf);
            memcpy(data, &apdu[i * max_data_len], num_bytes);

            // Move on to next packet in the packet chain
            curr_pkt = curr_pkt->next;
        }

        // Enqueue the packet(s) onto the packet queue
        tera_pkt_q_enqueue(pkt_q, pkt);
    } // end if

    return ret;
} // scp_pkt_send()

// --------------------------------------------------------------------------
//  tera_scp_init
///     Initialize the SCP module
//
//  Description:
///     This function initializes the SCP module by initializing the variables
///     in the control block.
///
/// \return
///     TERA_SUCCESS
///
INT32 tera_scp_init(
    struct pci_dev *dev     // Device object
    )
{
    sTERA_DEV_CBLK *cblk = pci_get_drvdata(dev);

    // Mark the module as initialized
    cblk->scp_cblk.initialized = 1;
    cblk->scp_cblk.instances = NULL;

    return TERA_SUCCESS;
} // tera_scp_init()

// --------------------------------------------------------------------------
//  tera_scp_cleanup
///     Clean up the SCP module
//
//  Description:
///     This function cleans up the SCP module. This function marks the
///     module as uninitialized, and deletes all instances of the SCP. The
///     user must NOT use any SCP instances anymore after calling this function.
///
/// \return
///     None
///
void tera_scp_cleanup(
    struct pci_dev *dev     ///< Device object
    )
{
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sSCP_CBLK      *cblk = &dev_cblk->scp_cblk;
    sSCP_INSTANCE  *instance;

    mTERA_LOG_DBG("start");

    if (cblk->initialized)
    {
        // Mark the module as uninitialized
        cblk->initialized = 0;

        // Free all memory used by all the instances
        while (cblk->instances != NULL)
        {
            instance = cblk->instances;
            cblk->instances = instance->next;
            kfree(instance);
        }
    }

    mTERA_LOG_DBG("end");
} // tera_scp_cleanup()

// --------------------------------------------------------------------------
//  tera_scp_create
///     Create a SCP instance
//
//  Description:
///     This function creates a SCP instance. Each instance can handle up to
///     TERA_SCP_NUM_CHAN channels.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_scp_create(
    struct pci_dev           *dev,              ///< Device object
    UINT8                     active_chan,      ///< Active SCP channel IDs (bitmapped)
    TERA_SCP_HANDLE          *handle,           ///< SCP handle
    fTERA_SCP_CHAN_OPEN_CBACK open_cback,       ///< Channel open callback
    void                     *open_cback_arg    ///< Callback argument
    )
{
    INT32           ret = TERA_ERR_FAILURE;
    sTERA_DEV_CBLK *dev_cblk = pci_get_drvdata(dev);
    sSCP_CBLK      *cblk = &dev_cblk->scp_cblk;
    sSCP_INSTANCE  *instance;
    sSCP_INSTANCE  *node;

    ASSERT(handle != NULL);

    if (cblk->initialized)
    {
        instance = kzalloc(sizeof(sSCP_INSTANCE), GFP_KERNEL);
        if (instance == NULL)
        {
            mTERA_LOG_ERR("failed allocate memory for SCP instance");
        }
        else
        {
            memset(instance, 0, sizeof(sSCP_INSTANCE));

            // Copy down the active SCP channel IDs and channel open callback
            instance->active = active_chan;
            instance->open_cback = open_cback;
            instance->open_cback_arg = open_cback_arg;

            // Set the MTU to the default size
            instance->mtu = TERA_SCP_DEFAULT_MTU;

            //
            // Add the instance to the list in the control block so we can
            // free the allocated memory during cleanup
            //
            if (cblk->instances == NULL)
            {
                cblk->instances = instance;
            }
            else
            {
                node = cblk->instances;

                for (;;)
                {
                    if (node->next == NULL)
                    {
                        node->next = instance;
                        break;
                    }

                    node = node->next;
                }
            }

            *handle = instance;
            ret = TERA_SUCCESS;
        } // end if
    } // end if

    return ret;
} // tera_scp_create()

// --------------------------------------------------------------------------
//  tera_scp_reset
///     Reset a SCP instance
//
//  Description:
///     This function resets a SCP instance and puts all channels into the
///     reset state. Since there is no BYE SCP signoff mechanism, the SCP does
///     not know when to reset the channels, and it is up to the user of the
///     SCP to call this function whenever the communication path is torn
///     down.
///
/// \return
///     None
///
void tera_scp_reset(
    TERA_SCP_HANDLE handle
    )
{
    sSCP_INSTANCE *instance = handle;

    mTERA_LOG_DBG("start");

    ASSERT(handle != NULL);

    // Clear the ready flags for all the channels
    instance->ready = 0;
} // tera_scp_reset()

// --------------------------------------------------------------------------
//  tera_scp_mtu_set
///     Set the MTU for a SCP instance
//
//  Description:
///     This function sets the MTU for a SCP instance. This function is optional
///     and is only needed if the desired MTU is different than the default
///     size (see TERA_SCP_DEFAULT_MTU). The MTU size is used by the SCP to
///     segment the outbound packets into buffers that are equal to or smaller
///     than the MTU. If the MTU is less than the size of the SCP and SAR
///     header plus 1 byte, then this function returns TERA_ERR_FAILURE and the
///     MTU is not changed.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_scp_mtu_set(
    TERA_SCP_HANDLE handle,
    UINT16          mtu
    )
{
    INT32          ret = TERA_ERR_FAILURE;
    sSCP_INSTANCE *instance = handle;

    ASSERT(handle != NULL);

    if (mtu <= (sizeof(sSCP_PKT_HDR) + sizeof(sSAR_PKT_HDR) + 1))
    {
        mTERA_LOG_ERR("mtu too small");
    }
    else
    {
        instance->mtu = mtu;

        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_scp_mtu_set()

// --------------------------------------------------------------------------
//  tera_scp_chan_is_ready
///     Check if the SCP channel is ready
//
//  Description:
///     This function checks if the specified SCP channel is ready for
///     transmission, i.e., if the SCP_INVITE was received. If it is ready,
///     this function returns 1, else 0.
///
/// \return
///     1,
///     0
///
INT tera_scp_chan_is_ready(
    TERA_SCP_HANDLE handle,
    UINT8           chan_id
    )
{
    sSCP_INSTANCE *instance = handle;

    ASSERT(handle != NULL);

    return (instance->ready & (1 << chan_id)) ? 1 : 0;
} // tera_scp_chan_is_ready()

// --------------------------------------------------------------------------
//  tera_scp_pkt_proc
///     Process a SCP packet
//
//  Description:
///     This function processes a SCP packet. This function looks at the SCP
///     packet in the packet buffer. Only packets that are meant for channels
///     that are active are processed. This funciton does NOT free the buffer
///     memory. It is up to the caller to do that.
///
///     If the function receives a SCP INVITE packet, a SCP OK packet is sent
///     back to open the SCP channel. The SCP OK packet is enqueued onto the
///     specified queue.
///
///     If its a SCP DATA packet, the function will point the APDU pointer to
///     the part of the buffer where the APDU begins. The channel ID and APDU
///     length is also filled out.
///
///     In other words, the caller of this function should look at the APDU
///     pointer and carry out further processing if the process the APDU if the
///     APDU pointer is non-null. In all other cases, the caller should just
///     free the buffer memory.
///
/// \return
///     None
///
void tera_scp_pkt_proc(
    struct pci_dev *dev,        ///< Device object
    TERA_SCP_HANDLE handle,     ///< SCP handle
    sTERA_MEM_BUF  *buf,        ///< Packet buffer
    UINT32          buf_len,    ///< Packet buffer length
    UINT8          *chan_id,    ///< Channel ID of packet
    UINT8         **apdu,       ///< Pointer to APDU in packet
    UINT32         *apdu_len,   ///< APDU length
    sTERA_PKT_Q    *pkt_q       ///< Packet queue for sending SCP OK packets
    )
{
    sSCP_INSTANCE *instance = handle;
    sSCP_PKT_HDR  *scp_pkt_hdr = mSCP_PKT_HDR_PTR_GET(buf);
    sSAR_PKT_HDR  *sar_pkt_hdr = mSAR_PKT_HDR_PTR_GET(buf);
#if 0
    UINT8         *pkt_data    = mDATA_PTR_GET(buf);
    UINT32         i;
#endif

    ASSERT(handle != NULL);
    ASSERT((buf_len == 0) || (buf != NULL));
    ASSERT(chan_id != NULL);
    ASSERT(apdu != NULL);
    ASSERT(apdu_len != NULL);
    ASSERT(pkt_q != NULL);

    *apdu = NULL;

    if (buf_len < (sizeof(sSCP_PKT_HDR) + sizeof(sSAR_PKT_HDR)))
    {
        mTERA_LOG_ERR("invalid SCP packet length (buf_len=%u)", buf_len);
    }
    else if (!(instance->active & (1 << scp_pkt_hdr->chan_id)))
    {
        mTERA_LOG_DBG("SCP packet on inactive channel (chan_id=%u instance->active=0x%X)",
                      scp_pkt_hdr->chan_id,
                      instance->active);
    }
    else if (sar_pkt_hdr->apdu_seg_loc != APDU_SEG_LOC_ONLY_ONE)
    {
        //
        // We currently do not support reassembly so drop all non-single
        // packets. Under normal conditions this should not happen. One
        // customer reported this message spammed their log file after
        // the PCoIP host card reset while the workstation was running.
        // This is not a recoverable error and the workstation must be
        // reset, but spamming the log file is bad, which is why this
        // message has been changed to a debug message, which is not
        // output by default.
        //
        mTERA_LOG_DBG("dropping segmented SCP packet");
    }
    else
    {
#if 0
        mTERA_LOG_DBG("scp: chan_id=%u pkt_type=%u buf_len=%u",
                      scp_pkt_hdr->chan_id, scp_pkt_hdr->pkt_type,
                      buf_len);
        mTERA_LOG_DBG("sar: apdu_size=%u apdu_seg_size=%u apdu_seg_loc=%u",
                      mTERA_UINT32_BYTE_SWAP(sar_pkt_hdr->apdu_size),
                      mTERA_UINT16_BYTE_SWAP(sar_pkt_hdr->apdu_seg_size),
                      sar_pkt_hdr->apdu_seg_loc);
        for (i=0; i<buf_len; i=i+16)
        {
            if((buf_len-i) > 15)
            {
                mTERA_LOG_DBG("data[%03d]: %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X",
                              i,
                              pkt_data[i],   pkt_data[i+1], pkt_data[i+2], pkt_data[i+3],
                              pkt_data[i+4], pkt_data[i+5], pkt_data[i+6], pkt_data[i+7],
                              pkt_data[i+8], pkt_data[i+9], pkt_data[i+10], pkt_data[i+11],
                              pkt_data[i+12], pkt_data[i+13], pkt_data[i+14], pkt_data[i+15]);
            }
            else if((buf_len-i) > 11)
            {
                mTERA_LOG_DBG("data[%03d]: %02X%02X%02X%02X %02X%02X%02X%02X %02X%02X%02X%02X",
                              i,
                              pkt_data[i],   pkt_data[i+1], pkt_data[i+2], pkt_data[i+3],
                              pkt_data[i+4], pkt_data[i+5], pkt_data[i+6], pkt_data[i+7],
                              pkt_data[i+8], pkt_data[i+9], pkt_data[i+10], pkt_data[i+11]);
            }
            else if((buf_len-i) > 7)
            {
                mTERA_LOG_DBG("data[%03d]: %02X%02X%02X%02X %02X%02X%02X%02X",
                              i,
                              pkt_data[i],   pkt_data[i+1], pkt_data[i+2], pkt_data[i+3],
                              pkt_data[i+4], pkt_data[i+5], pkt_data[i+6], pkt_data[i+7]);
            }
            else
            {
                mTERA_LOG_DBG("data[%03d]: %02X%02X%02X%02X",
                              i,
                              pkt_data[i],   pkt_data[i+1], pkt_data[i+2], pkt_data[i+3]);
            }
        }
#endif

        switch (scp_pkt_hdr->pkt_type)
        {
            case SCP_PKT_TYPE_INVITE:
            {
                mTERA_LOG_DBG("SCP INVITE packet, chan id=%d", scp_pkt_hdr->chan_id);

                // Send a SCP OK packet to start the channel
                scp_ok_pkt_send(dev, instance, scp_pkt_hdr->chan_id, pkt_q);

                if (instance->open_cback != NULL)
                {
                    instance->open_cback(scp_pkt_hdr->chan_id,
                                         instance->open_cback_arg);
                }

                break;
            }
            case SCP_PKT_TYPE_DATA:
            {
//                mTERA_LOG_DBG("SCP DATA packet");

                // Only process data packets if the channel is ready
                if (instance->ready & (1 << scp_pkt_hdr->chan_id))
                {
                    // Fill out the APDU pointer, APDU length, and channel ID
                    *apdu = mDATA_PTR_GET(buf);
                    *apdu_len = mTERA_UINT32_BYTE_SWAP(sar_pkt_hdr->apdu_size);
                    *chan_id = scp_pkt_hdr->chan_id;
                }
                else
                {
                    mTERA_LOG_DBG("stale SCP DATA packet (channel not ready)");
                }

                break;
            }
            case SCP_PKT_TYPE_OK:
            {
                mTERA_LOG_ERR("Rcvd unexpected SCP OK packet!!!");

                //
                // We should never get this packet because we never send out a
                // SCP INVITE packet
                //
                break;
            }
            default:
            {
                mTERA_LOG_ERR("invalid SCP packet type=%u",
                              scp_pkt_hdr->pkt_type);
                break;
            }
        }
    } // end if
} // tera_scp_pkt_proc()

// --------------------------------------------------------------------------
//  tera_scp_data_pkt_send
///     Send a SCP DATA packet
//
//  Description:
///     This function sends a SCP DATA packet with the specified APDU. If the
///     APDU is too large to fit into a single packet, the APDU is segmented
///     over multiple packets. This function fills in the SCP and SAR headers
///     of each packet. If successful, the packet is enqueued onto the packet
///     queue.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NULL_PTR,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_NO_MEMORY
///
INT32 tera_scp_data_pkt_send(
    struct pci_dev *dev,
    TERA_SCP_HANDLE handle,
    UINT8           chan_id,
    sTERA_PKT_Q    *pkt_q,
    void           *apdu,
    UINT32          apdu_size
    )
{
    INT32          ret;
    sSCP_INSTANCE *instance = handle;

    ASSERT(handle != NULL);
    ASSERT(chan_id < TERA_SCP_NUM_CHAN);
    ASSERT(pkt_q != NULL);

    if ((apdu_size != 0) && (apdu == NULL))
    {
        mTERA_LOG_ERR("got NULL buffer pointer for non-zero buffer length");

        ret = TERA_ERR_NULL_PTR;
    }
    else if (!(instance->ready & (1 << chan_id)))
    {
        mTERA_LOG_ERR("SCP channel %u not ready, instance->ready=%d",
                      chan_id, instance->ready);

        ret = TERA_ERR_INVALID_STATE;
    }
    else
    {
        // Send the SCP DATA packet(s)
        ret = scp_pkt_send(dev, chan_id, instance->mtu, pkt_q,
                           SCP_PKT_TYPE_DATA, apdu, apdu_size);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("failed to send SCP DATA packet(s) (ret=%d)", ret);
        }
    }

    return ret;
} // tera_scp_data_pkt_send()


