////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     T/L/V functions
///
///     The Type/Length/Value (T/L/V) library provides routines to set
///     and get a T/L/V tuple in a byte stream.  T/L/V can be used to
///     store a typed parameter in a byte stream, in binary format.
///     The Length field allows T/L/V parsers to skip over unknown
///     parameter Types.
///
///     This file is a generic T/L/V utility based loosely on
///     Interchange File Format (IFF), except does not support
///     FORM, LIST and CAT or other constructs found in IFF or RIFF.
///     This utility is not written to a standard specification,
///     thus it is just called "tlv" instead of "tlv_iff", etc.
///
///     The T/L/V field definitions are as follows:
///     - Type:
///         32-bit field uniquely identifies the Value field content.
///         Four Character Code (FourCC) consisting of 1 to 4 ASCII
///         alphanumeric characters, padded on the right with blank
///         characters (ASCI character code 32).  The FourCC is stored
///         as 4-bytes in memory and is conveniently used in 32-bit
///         comparisons.  For example, 'FOO ' is stored as 'F', 'O', 'O',
///         and ' ' (space, ASCII character value 32) in ascending
///         addresses in memory.  Stored in network byte order in
///         byte stream.
///
///     - Length:
///         32-bit field that identifies the number of bytes in
///         the Value field, excluding pad bytes at end of Value.
///         Stored in network byte order in byte stream.
///
///     - Value:
///         One or more bytes of binary user data, starting on a
///         32-bit aligned address.  If Length is not a multiple of
///         4, pad bytes with value 0x00 are written after data.
///
///     The library offers set/get for T/L/V tuples.  The library
///     provides several types of functions depending on the Value
///     field size:
///     - 8-bit Value (with 3 pad bytes)
///     - 16-bit Value (stored in network byte order, with 2 pad bytes)
///     - 32-bit Value (stored in network byte order)
///     - raw byte array Value (with 0, 1, 2, 3 pad bytes)
//
////////////////////////////////////////////////////////////////////////////

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_bstream.h"
#include "tera_tlv.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// pad value to align the V field to a 32-bit boundary
#define TERA_TLV_PAD_VALUE      0x00

// maximum number of pad bytes in Value field
#define TERA_TLV_MAX_PAD_BYTES  3

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  tlv_get_num_pad_bytes
//      Gets the number of pad bytes in Value field
//
//  Description:
//      This function uses the length field to calculate the number of
//      pad bytes in the Value field.  The Value field in the byte
//      stream is a multiple of 4-bytes (32-bit aligned).
//
//  \return
//      Number of pad bytes in Value field
//
static UINT8 tlv_get_num_pad_bytes(
    UINT32  length                      ///< number of bytes in Value field
                                        ///< excluding pad bytes
    )
{
    UINT8   non_aligned_bytes;
    UINT8   num_pad_bytes;


    // calculate the number of non-aligned bytes in Value field
    // (aligned to 32-bit boundary)
    non_aligned_bytes = length % sizeof(UINT32);

    if(non_aligned_bytes != 0)
    {
        // Value field has 1, 2, or 3 bytes outside of a 32-bit boundary.
        // Calculate the number of pad bytes required to align Value field.
        num_pad_bytes = sizeof(UINT32) - non_aligned_bytes;
    }
    else
    {
        // Value field is aligned to a 32-bit boundary.
        // No pad bytes required to align Value field.
        num_pad_bytes = 0;
    }

    return(num_pad_bytes);
}


// --------------------------------------------------------------------------
//  tera_tlv_set_tlvraw
///     Sets the T/L/V tuple for a raw byte array of values
//
//  Description:
///    This function stores the following in the byte stream:
///     - Type: FourCC in network byte order
///     - Length: number of bytes in Value field provided by user,
///       in network byte order (excluding pad bytes).
///     - Value: raw data bytes.  If the length of the raw data is
///       not a multiple of 4 bytes, this routine pads the Value
///       field so that the byte stream offset indexes the first
///       byte of the next T/L/V on a 32-bit aligned address.
///
/// \return
///     None
///
void tera_tlv_set_tlvraw(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                  type,       ///< Type field (FourCC)
    UINT32                  length,     ///< Length field (bytes in value array)
    UINT8                  *value       ///< Pointer to array of bytes
    )
{
    UINT8   num_pad_bytes;
    UINT8   pad[3];
    UINT8   i;

    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(length != 0);
    ASSERT(value != NULL);

    // Type field - must be in network byte order in byte stream
    tera_bstream_put32(cb,
                       mTERA_UINT32_BYTE_SWAP(type));

    // Length field - must be in network byte order in byte stream
    tera_bstream_put32(cb,
                       mTERA_UINT32_BYTE_SWAP(length));

    // Value field - copy array of bytes provided by user
    tera_bstream_putraw(cb,
                        length,
                        value);

    // calculate the number of pad bytes for Value field
    num_pad_bytes = tlv_get_num_pad_bytes(length);

    if(num_pad_bytes != 0)
    {
        // pad to make number of bytes in Value field aligned to
        // a 32-bit boundary
        for(i=0; i<num_pad_bytes; i++)
        {
            pad[i] = TERA_TLV_PAD_VALUE;
        }

        // store pad bytes in byte stream
        tera_bstream_putraw(cb,
                            num_pad_bytes,
                            &pad[0]);
    }
}


// --------------------------------------------------------------------------
//  tera_tlv_set_tlv8
///     Sets the T/L/V tuple for a 8-bit value
//
//  Description:
///     This function stores the Type, Length, and Value fields including
///     3-bytes of pad.  After calling this routine, the byte stream offset
///     indexes the first byte of the next T/L/V.
///
/// \return
///     None
///
void tera_tlv_set_tlv8(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                  type,       ///< Type field (FourCC)
    UINT8                   value       ///< Value field (8-bit value)
    )
{
    UINT8 tmp;


    tmp = value;

    // store T/L/V for 8-bit value in byte stream
    tera_tlv_set_tlvraw(cb,
                        type,
                        sizeof(UINT8),
                        &tmp);
}

// --------------------------------------------------------------------------
//  tera_tlv_set_tlv16
///     Sets the T/L/V tuple for a 16-bit value
//
//  Description:
///     This function converts the 16-bit value from host byte order
///     to network byte order and stores the Type, Length, and Value
///     fields including 2-bytes of pad.  After calling this routine,
///     the byte stream offset indexes the first byte of the next T/L/V.
///
/// \return
///     None
///
void tera_tlv_set_tlv16(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                  type,       ///< Type field (FourCC)
    UINT16                  value       ///< Value field (16-bit value)
    )
{
    UINT16 tmp;


    // convert from host byte order to network byte order
    tmp = mTERA_UINT16_BYTE_SWAP(value);

    // store T/L/V for 16-bit value in byte stream
    tera_tlv_set_tlvraw(cb,
                        type,
                        sizeof(UINT16),
                        (UINT8 *)&tmp);
}

// --------------------------------------------------------------------------
//  tera_tlv_set_tlv32
///     Sets the T/L/V tuple for a 32-bit value
//
//  Description:
///     This function converts the 32-bit value from host byte order
///     to network byte order and stores the Type, Length, and Value
///     fields.  After calling this routine, the byte stream offset
///     indexes the first byte of the next T/L/V.
///
/// \return
///     None
///
void tera_tlv_set_tlv32(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                  type,       ///< Type field (FourCC)
    UINT32                  value       ///< Value field (32-bit value)
    )
{
    UINT32 tmp;


    // convert 32-bit value from host byte order to network byte order
    tmp = mTERA_UINT32_BYTE_SWAP(value);

    // store T/L/V for 32-bit value in byte stream
    tera_tlv_set_tlvraw(cb,
                        type,
                        sizeof(UINT32),
                        (UINT8 *)&tmp);
}


// --------------------------------------------------------------------------
//  tera_tlv_set_tlvlldp
///     Sets the T/L/V tuple for an LLDP format byte array of values
//
//  Description:
///    This function stores the following in the byte stream:
///     - Type: 7 bit LLDP type code
///     - Length: number of bytes in Value field provided by user
///     - Value: raw data bytes.  32 bit alignment is not guaranteed
///         as a valid length is 0
///
///     The TLV is contructed as described in IEEE 802.1AB-2009 section 8.4.
///
/// \return
///     None
///
void tera_tlv_set_tlvlldp(
    sTERA_BSTREAM_CBLK      *cb,        ///< pointer to byte stream cblk
    UINT32                  type,       ///< Type field (7 bit LLDP code)
    UINT32                  length,     ///< Length field (bytes in value array)
    UINT8                   *value      ///< Pointer to array of bytes
    )
{
    UINT16  tlv_header;

    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(value != NULL);

    // construct the first short from the type and length
    tlv_header =  (UINT16)  (((UINT8)type << TERA_TLV_LLDP_TYPE_BITOFFSET) & TERA_TLV_LLDP_TYPE_MASK)
                          | ((UINT16)length & TERA_TLV_LLDP_LENGTH_MASK);

    // TLV header field - must be in network byte order in byte stream
    tera_bstream_put16(cb,
                       mTERA_UINT16_BYTE_SWAP(tlv_header));

    if(length != 0)
    {
        // Value field - copy array of bytes provided by user
        tera_bstream_putraw(cb,
                            length,
                            value);
    }

    return;
}


// --------------------------------------------------------------------------
//  tera_tlv_get_tl
///     Gets the Type and Length fields from the T/L/V tuple
//
//  Description:
///     This function reads the Type (FourCC) and Length fields from a T/L/V.
///     The routine handles the network byte order swap for the Type
///     and Length fields.  The function returns the Type, Length,
///     and the offset to the next T/L/V tuple in the byte stream.
///     If the Type is known to the caller, the caller should call
///     tera_tlv_get_xxx routines to get the Value field from the T/L/V.
///     If the Type is unknown to the caller, the caller can call
///     tera_bstream_set_curpos with the value stored by next_tlv_offset
///     in order to advance the byte stream offset to the next T/L/V
///     tuple.  This function leaves the byte stream offset at the first
///     byte in the Value field of the current T/L/V.
///
/// \return
///     None
///
void tera_tlv_get_tl(
    sTERA_BSTREAM_CBLK *cb,             ///< pointer to byte stream cblk
    UINT32             *type,           ///< pointer to destination to return
                                        ///< Type (FourCC)
    UINT32             *length,         ///< pointer to destination to return
                                        ///< Length
    UINT32             *next_tlv_offset ///< pointer to destination to return byte
                                        ///< stream offset to next T/L/V tuple
    )
{
    UINT8   num_pad_bytes;
    UINT32  next_offset;
    UINT32  tlv_type;
    UINT32  tlv_length;

    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(type != NULL);
    ASSERT(length != NULL);
    ASSERT(next_tlv_offset != NULL);

    // get Type field from byte stream
    tlv_type = tera_bstream_get32(cb);

    // convert Type from network byte order to host byte order
    tlv_type = mTERA_UINT32_BYTE_SWAP(tlv_type);

    // get Length field from byte stream
    tlv_length = tera_bstream_get32(cb);

    // convert Length from network byte order to host byte order
    tlv_length = mTERA_UINT32_BYTE_SWAP(tlv_length);

    // get the number of pad bytes in Value field
    num_pad_bytes = tlv_get_num_pad_bytes(tlv_length);

    // offset to start of next T/L/V tuple in byte stream (current byte stream
    // offset indexes the first byte of Value field)
    next_offset = tera_bstream_get_curpos(cb) + tlv_length + num_pad_bytes;

    // return Type and Length for current T/L/V
    *type = tlv_type;
    *length = tlv_length;

    // return byte stream offset to next T/L/V
    *next_tlv_offset = next_offset;
}


// --------------------------------------------------------------------------
//  tera_tlv_get_tl_lldp
///     Gets the Type and Length fields from the T/L/V tuple that is in LLDP
///     format.
//
//  Description:
///     This function reads the Type (7 bit LLDP type) and Length fields
///     from a T/L/V.
///     The routine handles the network byte order swap for the Type
///     and Length fields.  The function returns the Type, Length,
///     and the offset to the next T/L/V tuple in the byte stream.
///     If the Type is known to the caller, the caller should call
///     tera_tlv_get_xxx routines to get the Value field from the T/L/V.
///     If the Type is unknown to the caller, the caller can call
///     tera_bstream_set_curpos with the value stored by next_tlv_offset
///     in order to advance the byte stream offset to the next T/L/V
///     tuple.  This function leaves the byte stream offset at the first
///     byte in the Value field of the current T/L/V.
///
/// \return
///     None
///
void tera_tlv_get_tl_lldp(
    sTERA_BSTREAM_CBLK *cb,             ///< pointer to byte stream cblk
    UINT32             *type,           ///< pointer to destination to return
                                        ///< Type
    UINT32             *length,         ///< pointer to destination to return
                                        ///< Length
    UINT32             *next_tlv_offset ///< pointer to destination to return byte
                                        ///< stream offset to next T/L/V tuple
    )
{
    UINT16  tlv_header;

    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(type != NULL);
    ASSERT(length != NULL);
    ASSERT(next_tlv_offset != NULL);

    // get Type/Length fields from byte stream
    tlv_header = tera_bstream_get16(cb);

    // convert Type from network byte order to host byte order
    tlv_header = mTERA_UINT16_BYTE_SWAP(tlv_header);

    // get Length field from header
    *type = (UINT32)((tlv_header & TERA_TLV_LLDP_TYPE_MASK) >> TERA_TLV_LLDP_TYPE_BITOFFSET);
    *length = (UINT32)(tlv_header & TERA_TLV_LLDP_LENGTH_MASK);

    // offset to start of next T/L/V tuple in byte stream (current byte stream
    // offset indexes the first byte of Value field)
    *next_tlv_offset = tera_bstream_get_curpos(cb) + *length;
}


// --------------------------------------------------------------------------
//  tera_tlv_get_vraw
///     Gets raw bytes from the Value field of a T/L/V tuple
//
//  Description:
///     This function reads a raw byte array from the Value field of
///     a T/L/V.  The function assumes that the byte stream offset
///     currently indexes the Value field (assume that the caller
///     has called tera_tlv_get_tl before calling this function).
///     The function returns with the byte stream offset indexing
///     the Type field of the next T/L/V tuple in the byte stream.
///
/// \return
///     None
///
void tera_tlv_get_vraw(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                  length,     ///< number of raw bytes to read
    UINT8                  *value       ///< pointer to destination to store
                                        ///< array of bytes
    )
{
    UINT8   num_pad_bytes;

    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(length != 0);
    ASSERT(value != NULL);

    // Value - get raw bytes from byte stream, store in user provided buffer
    tera_bstream_getraw(cb,
                        length,
                        value);

    // check if the Value field has pad bytes
    num_pad_bytes = tlv_get_num_pad_bytes(length);

    if(num_pad_bytes != 0)
    {
        // advance past pad bytes so that we leave the
        // byte stream offset indexing the first byte of next
        // T/L/V tuple
        tera_bstream_set_curpos(cb,
                                tera_bstream_get_curpos(cb)+num_pad_bytes);
    }
}


// --------------------------------------------------------------------------
//  tera_tlv_get_vraw_nopad
///     Gets raw bytes from the Value field of a T/L/V tuple
//
//  Description:
///     This function reads a raw byte array from the Value field of
///     a T/L/V.  The function assumes that the byte stream offset
///     currently indexes the Value field (assume that the caller
///     has called tera_tlv_get_tl before calling this function).
///     The function returns with the byte stream offset indexing
///     the Type field of the next T/L/V tuple in the byte stream.
///
/// \return
///     None
///
void tera_tlv_get_vraw_nopad(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                  length,     ///< number of raw bytes to read
    UINT8                  *value       ///< pointer to destination to store
                                        ///< array of bytes
    )
{
    // consistency checks
    ASSERT(cb != NULL);
    ASSERT(length != 0);
    ASSERT(value != NULL);

    // Value - get raw bytes from byte stream, store in user provided buffer
    tera_bstream_getraw(cb,
                        length,
                        value);
}


// --------------------------------------------------------------------------
//  tera_tlv_get_v8
///     Gets 8-bit Value field of a T/L/V tuple
//
//  Description:
///     This function reads a 16-bit Value field of a T/L/V tuple.
///     The function assumes that the byte stream offset
///     currently indexes the Value field (assume that the caller
///     has called tera_tlv_get_tl before calling this function).
///     The function returns with the byte stream offset indexing
///     the Type field of the next T/L/V tuple in the byte stream.
///
/// \return
///     None
///
void tera_tlv_get_v8(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT8                  *value       ///< pointer to destination to store
                                        ///< 8-bit value
    )
{

    // get 8-bit Value from T/L/V
    tera_tlv_get_vraw(cb,
                      sizeof(UINT8),
                      value);
}

// --------------------------------------------------------------------------
//  tera_tlv_get_v16
///     Gets 16-bit Value field of a T/L/V tuple
//
//  Description:
///     This function reads a 16-bit Value field of a T/L/V tuple.
///     The function assumes that the byte stream offset
///     currently indexes the Value field (assume that the caller
///     has called tera_tlv_get_tl before calling this function).
///     The function converts the 16-bit value from network byte order
///     to host byte order.  The function returns with the byte stream
///     offset indexing the Type field of the next T/L/V tuple in the
///     byte stream.
///
/// \return
///     None
///
void tera_tlv_get_v16(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT16                 *value       ///< pointer to destination to store
                                        ///< 16-bit value
    )
{
    UINT16 tmp;


    // get 16-bit Value from T/L/V
    tera_tlv_get_vraw(cb,
                      sizeof(UINT16),
                      (UINT8 *)&tmp);

    // convert from network byte order to host byte order
    *value = mTERA_UINT16_BYTE_SWAP(tmp);
}

// --------------------------------------------------------------------------
//  tera_tlv_get_v32
///     Gets 32-bit Value field of a T/L/V tuple
//
//  Description:
///     This function reads a 32-bit Value field of a T/L/V tuple.
///     The function assumes that the byte stream offset
///     currently indexes the Value field (assume that the caller
///     has called tera_tlv_get_tl before calling this function).
///     The function converts the 32-bit value from network byte order
///     to host byte order.  The function returns with the byte stream
///     offset indexing the Type field of the next T/L/V tuple in the
///     byte stream.
///
/// \return
///     None
///
void tera_tlv_get_v32(
    sTERA_BSTREAM_CBLK     *cb,         ///< pointer to byte stream cblk
    UINT32                 *value       ///< pointer to destination to store
                                        ///< 32-bit value
    )
{
    UINT32 tmp;


    // get 32-bit Value from T/L/V
    tera_tlv_get_vraw(cb,
                      sizeof(UINT32),
                      (UINT8 *)&tmp);

    // convert from network byte order to host byte order
    *value = mTERA_UINT32_BYTE_SWAP(tmp);
}

// --------------------------------------------------------------------------
//  tera_tlv_log_t
///     Logs a FourCC type code
//
//  Description:
///     This function takes a FourCC type and outputs it to the
///     event log.  This function assumes that the caller passes
///     in the Type in host byte order.  The function converts
///     the type to network byte order and logs to the event module.
///
/// \return
///     None
///
void tera_tlv_log_t(
    UINT32  type                        ///< FourCC Type
    )
{
    UINT32  net_type;
    char    *type_str;


    // convert type from host byte order to network byte order
    net_type = mTERA_UINT32_BYTE_SWAP(type);

    type_str = (char *)&net_type;

    mTERA_LOG_DBG("FourCC Type: %c%c%c%c",
                  type_str[0],
                  type_str[1],
                  type_str[2],
                  type_str[3]);
}

