////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Monitor Property Page Class Implementation
///
///     This file implements the PCoIP Host Software Monitor property page class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include <wx/display.h>

#include "tera_errors.h"
#include "tera_log.h"
#include "tera_types.h"
#include "tera_mem.h"
#include "tera_pcoip_agent_internal.h"
#include "tera_pcoip_agent_gui_prop_mon.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

#define NAME_COLUMN             0
#define VENDOR_ID_COLUMN        1
#define PRODUCT_ID_COLUMN       2
#define SERIAL_NUMBER_COLUMN    3
#define RESOLUTION_COLUMN       4

BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI_PROP_MON, wxPanel)
    EVT_INIT_DIALOG(cPCOIP_AGENT_GUI_PROP_MON::evt_init_dialog)
    EVT_BUTTON(wxID_APPLY_BTN, cPCOIP_AGENT_GUI_PROP_MON::evt_apply_btn)
    EVT_PG_CHANGED(wxID_ANY, cPCOIP_AGENT_GUI_PROP_MON::evt_pg_changed)
END_EVENT_TABLE()

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::evt_init_dialog
//      Process the evt_init_dialog event
//
//  Description:
//      This function processes the init dialog event. This function
//      call updates the property grid with the most up-to-date monitor
//      configuration.
//
void cPCOIP_AGENT_GUI_PROP_MON::evt_init_dialog(
    wxInitDialogEvent& evt
    )
{
    mTERA_LOG_DBG("MON EVT_INIT_DIALOG");

    // Populate the property grid with the current values
    prop_grid_populate(0);

    // Initially enable the Apply button. If any of the screen settings are not
    // configured properly disable the button.
    m_apply_btn->Enable(1);

    //
    // Look at the monitor information and check to see if it is invalid. If
    // it is update the user configured screen settings on the GUI. If this is
    // not done stale data is displayed on the GUI.
    //
    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        if(m_screen_list[i].mon == 255)
        {
            prop_grid_populate_screen_entry(i);
            m_apply_btn->Enable(0);
        }
    }

    evt.Skip();
} // cPCOIP_AGENT_GUI_PROP_MON::evt_init_dialog()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::evt_apply_btn
//      Process the apply button click event
//
//  Description:
//      This function calls the function that saves and applies the current
//      settings.
//
void cPCOIP_AGENT_GUI_PROP_MON::evt_apply_btn(
    wxCommandEvent& evt
    )
{
    int mon_info_ok;

    //
    // delete the monitor information for the attached client, this ensures
    // the data from the grid is used and written to the config file
    //
    delete_mon_usr_cfg();

    // clear flag indicating valid topology data has been sent to the client
    wxGetApp().gui().set_topo_sent_status(0);

    // Populate the property grid without reading the config file
    prop_grid_populate(1);

    // update the configuration file
    save_config();

    // Populate the property grid again, this time use the data in the config file
    prop_grid_populate(0);

    // send the latest monitor information to the FW
    mon_info_ok = wxGetApp().gui().display_topology_send();

    // send the latest settings to the client
    wxGetApp().gui().lcur_lkbd_settings_send(mon_info_ok);

    m_apply_btn->Enable(0);
} // cPCOIP_AGENT_GUI_PROP_MON::evt_apply_btn()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::evt_pg_changed
//      Process the page change event
//
//  Description:
//      This function processes the page change event. It finds the property
//      that was changed and updates the internal structures that need to
//      change.
//
void cPCOIP_AGENT_GUI_PROP_MON::evt_pg_changed(
    wxPropertyGridEvent& evt
    )
{
    wxString label;
    wxString val;
    UINT32   monitor;

    mTERA_LOG_DBG("EVT_PG_CHANGED");
    dump_dbg_info();

    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        label.Printf(_T("screen%u_mon"), i);

        if (evt.GetPropertyName() == label)
        {
            val = m_prop_grid->GetPropertyValueAsString(label);
            sscanf(val.char_str(), "Monitor %d", &monitor);

            mTERA_LOG_DBG("%s changed: screen %d, monitor %d",
                          label.c_str(), i, monitor);

            // update the monitor value for the screen if the field is non-blank
            if((monitor > 0) && (monitor <= MAX_MON_COUNT))
            {
                m_screen_list[i].mon = monitor;
            }

            prop_grid_populate_screen_entry(i);
        }
    }
    m_apply_btn->Enable(1);

    mTERA_LOG_DBG("end");
    dump_dbg_info();
} // cPCOIP_AGENT_GUI_PROP_MON::evt_pg_changed()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::delete_mon_usr_cfg
//      Delete the monitor data from the configuration file
//
//  Description:
//      This function deletes the monitor data from the configuration file
//      if a session is active.
//
void cPCOIP_AGENT_GUI_PROP_MON::delete_mon_usr_cfg(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;

    // Get the session link status
    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    if (sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED)
    {
        // delete the monitor information for the attached client
        wxGetApp().gui().host_cfg().delete_mon_data(sess_link.mac, MAX_MON_COUNT);
    }
} // cPCOIP_AGENT_GUI_PROP_MON::delete_mon_usr_cfg()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::session_is_active
//      Check sesssion state
//
//  Description:
//      This function returns 1 if a session is active, otherwise it
//      returns 0.
//
int cPCOIP_AGENT_GUI_PROP_MON::session_is_active(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;

    // Get the session link status
    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    return (sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED) ? 1 : 0;

} // cPCOIP_AGENT_GUI_PROP_MON::session_is_active()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::dump_dbg_info
//      Dump debug information
//
//  Description:
//      This function dumps internal state information that may be useful
//      in debugging problems
//
void cPCOIP_AGENT_GUI_PROP_MON::dump_dbg_info(void)
{
    mTERA_LOG_DBG("Number of monitors: %d, Number of screens %d", m_mon_count, m_screen_count);

    for (UINT32 i = 0; i < MAX_MON_COUNT; i++)
    {
        // if a monitor is connected display the monitor data
        if (is_client_mon_connected(i))
        {
            mTERA_LOG_DBG("\tMonitor %d", i+1);
            mTERA_LOG_DBG("\t\tName:         %S", m_mon_list[i].name.c_str());
            mTERA_LOG_DBG("\t\tVendorID:     %S", m_mon_list[i].vid.c_str());
            mTERA_LOG_DBG("\t\tProductID:    %S", m_mon_list[i].pid.c_str());
            mTERA_LOG_DBG("\t\tSerialNumber: %S", m_mon_list[i].serial.c_str());
            mTERA_LOG_DBG("\t\tResolution:   %S", m_mon_list[i].resolution.c_str());
        }
    }

    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        mTERA_LOG_DBG("\tScreen[%d] %d - size %dx%d, pos (%d,%d)",
                      i, m_screen_list[i].mon,
                      m_screen_list[i].width, m_screen_list[i].height,
                      m_screen_list[i].pos_x, m_screen_list[i].pos_y);
    }
} // cPCOIP_AGENT_GUI_PROP_MON::dump_dbg_info()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::save_config
//      Save the current configuration data
//
//  Description:
//      This function saves the current configuration settings. It is called by
//      the apply button event handler. It is also called when the agent detects
//      there is only one monitor and screen present.
//
//      This function first validates that all values in the property grid.
//      If all values are valid, then the values are stored into the
//      user configuration. If the settings are valid, then the monitor
//      information is also sent to the firmware.
//
void cPCOIP_AGENT_GUI_PROP_MON::save_config(void)
{
    INT32         cfg_ok = 1;
    wxString      label;
    wxString      vid, pid, serial;
    wxString      vid2, pid2, serial2;

    mTERA_LOG_DBG("save config start: current state");
    dump_dbg_info();

    // Clear current selection, if any
    m_prop_grid->ClearSelection();

    //
    // do nothing if the number of monitors reported by FW is zero, this can
    // happen if the host is not running FW release 2.3 or higher
    //
    if (m_mon_count == 0)
    {
        mTERA_LOG_DBG("m_mon_count = 0");
    }
    else
    {
        for (UINT32 i = 0; i < (m_screen_count-1); i++)
        {
            mTERA_LOG_DBG("Screen %u - Monitor %u", i, m_screen_list[i].mon);

            // Verify the monitor value of other screens are unique
            for (UINT32 j = i + 1; j < m_screen_count; j++)
            {
                if (m_screen_list[i].mon == m_screen_list[j].mon)
                {
                    mTERA_LOG_DBG("Identical Monitor<=>Screen information found");

                    label.Printf(_T("screen%u_mon"), i);
                    m_prop_grid->SetPropertyTextColour(label, wxColour(255, 0, 0));
                    label.Printf(_T("screen%u_mon"), j);
                    m_prop_grid->SetPropertyTextColour(label, wxColour(255, 0, 0));
                    cfg_ok = 0;
                }
            }
        } // end for

        for (UINT32 i = 0; i < m_screen_count; i++)
        {
            if ((m_screen_list[i].mon == 0) || (m_screen_list[i].mon > MAX_MON_COUNT))
            {
                cfg_ok = 0;
                mTERA_LOG_DBG("Invalid monitor number %d in m_screeen_list[%d].mon field",
                              m_screen_list[i].mon, i);
            }
        }

        // the monitor information is only valid when a session is active
        if (!session_is_active())
        {
            cfg_ok = 0;
            mTERA_LOG_DBG("session inactive - do not save monitor info");
        }

        // Only save values if they are all valid
        if (cfg_ok)
        {
            wxString    usr_cfg_path;
            UINT32      monitor;
            sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;

            mTERA_LOG_DBG("saving monitor info");
            dump_dbg_info();

            // Get the session link status
            mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
            wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

            // Save to user configuration
            wxGetApp().gui().host_cfg().set_monitor_screen_counts(sess_link.mac,
                                                                  m_mon_count,
                                                                  m_screen_count);

            for (UINT32 i = 0; i < m_screen_count; i++)
            {
                if (TERA_SUCCESS == get_idx_monitor(m_screen_list[i].mon, &monitor))
                {
                    wxGetApp().gui().host_cfg().set_monitor_info(sess_link.mac,
                                                                 i+1,
                                                                 &m_mon_list[monitor].vid,
                                                                 &m_mon_list[monitor].pid,
                                                                 &m_mon_list[monitor].serial,
                                                                 m_screen_list[i].width,
                                                                 m_screen_list[i].height,
                                                                 m_screen_list[i].pos_x,
                                                                 m_screen_list[i].pos_y,
                                                                 m_screen_list[i].mon);
                }
                else
                {
                    mTERA_LOG_ERR("Error finding monitor %d in the list of monitors for screen %d",
                                  m_screen_list[i].mon, i);
                    break;
                }
            }
        }
        else
        {
            // Try to grab the user's attention by ringing the system bell
            ::wxBell();
        }
    }
    mTERA_LOG_DBG("End");
} // cPCOIP_AGENT_GUI_PROP_MON::save_config()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::mon_list_clear
//      Clear the internal monitor list
//
//  Description:
//      This function clears the internal monitor list and also resets the
//      internal monitor list count to zero.
//
void cPCOIP_AGENT_GUI_PROP_MON::mon_list_clear(void)
{
    for (UINT32 i = 0; i < 4; i++) //TODO Get rid of magic number
    {
        m_mon_list[i].name = wxEmptyString;
        m_mon_list[i].vid = wxEmptyString;
        m_mon_list[i].pid = wxEmptyString;
        m_mon_list[i].serial = wxEmptyString;
        m_mon_list[i].resolution = wxEmptyString;
    }

    m_mon_count = 0;
} // cPCOIP_AGENT_GUI_PROP_MON::mon_list_clear()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::mon_list_init
//      Clear and initialize the internal monitor list
//
//  Description:
//      This function initializes the internal monitor list by retrieving the
//      monitor information from the firmware. If a session is not active all
//      entries in the monitor list structure will be zero. If a session is
//      active the function will attempt to read the firmware information
//      multiple times. This is done because sometimes the FW will report no
//      monitors are connected immediately after a session is established.
//
void cPCOIP_AGENT_GUI_PROP_MON::mon_list_init(void)
{
    bool                             done = false;
    sTERA_PCOIP_LINK_CACHE_DISP_TOPO disp_topo;

    mon_list_clear();
    while (!done)
    {
        // the monitor information is only valid when a session is active
        if (!session_is_active())
        {
            done = true;
            mTERA_LOG_DBG("session inactive");
        }

        // if valid client topology data is available
        wxGetApp().gui().get_client_topo(&disp_topo);
        if (disp_topo.num_disp != 0)
        {
            m_mon_count = 0;

            for (UINT32 i = 0; i < TERA_PCOIP_HCC_APDU_MAX_NUM_DISP; i++)
            {
                mTERA_LOG_DBG("\tMonitor %u: name   %s", i, disp_topo.iden[i].name);
                mTERA_LOG_DBG("\tMonitor %u: vid    %s", i, disp_topo.iden[i].vid);
                mTERA_LOG_DBG("\tMonitor %u: pid    %s", i, disp_topo.iden[i].pid);
                mTERA_LOG_DBG("\tMonitor %u: serial %s", i, disp_topo.iden[i].serial);
                mTERA_LOG_DBG("\tMonitor %u: resolution %dx%d", i, disp_topo.disp[i].width, disp_topo.disp[i].height);

                // Add this into the list of monitors
                m_mon_list[i].name.Printf(_T("%S"), disp_topo.iden[i].name);
                m_mon_list[i].vid.Printf(_T("%S"), disp_topo.iden[i].vid);
                m_mon_list[i].pid.Printf(_T("%S"), disp_topo.iden[i].pid);
                m_mon_list[i].serial.Printf(_T("%S"), disp_topo.iden[i].serial);
                m_mon_list[i].resolution.Printf(_T("%dx%d"), disp_topo.disp[i].width, disp_topo.disp[i].height);

                // increment the monitor count if a monitor is attached
                if ((strlen(disp_topo.iden[i].name)) ||
                    (strlen(disp_topo.iden[i].vid))  ||
                    (strlen(disp_topo.iden[i].pid))  ||
                    (strlen(disp_topo.iden[i].serial)))
                {
                    m_mon_count++;
                }
            }
            mTERA_LOG_DBG("\tNumber of monitors = %d", m_mon_count);
            done = true;
        }
        else
        {
            // This can occur when the FW gets into a state where a message is
            // stuck in the imaging managers Q. One way to resolve this is to
            // reset the workstation.
            mTERA_LOG_DBG("Unable to get mon info from FW: cleared m_mon_list");
            done=true;
            mon_list_clear();
            m_cache_topo_read_err = 1;
        }
    }
} // cPCOIP_AGENT_GUI_PROP_MON::mon_list_init()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::screen_list_clear
//      Clear the internal screen list
//
//  Description:
//      This function clears the internal screen list and also resets the
//      internal screen list count to zero.
//
void cPCOIP_AGENT_GUI_PROP_MON::screen_list_clear(void)
{
    for (UINT32 i = 0; i < MAX_MON_COUNT; i++)
    {
        m_screen_list[i].width = 0;
        m_screen_list[i].height = 1;
        m_screen_list[i].pos_x = 0;
        m_screen_list[i].pos_y = 0;
        m_screen_list[i].mon = 255;
    }
    m_screen_count = 0;
} // cPCOIP_AGENT_GUI_PROP_MON::screen_list_clear()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::screen_list_init
//      Clear and initialize the internal screen list
//
//  Description:
//      This function initializes the internal screen list with information
//      provided by the caller. The information is stored in the screen list.
//
void cPCOIP_AGENT_GUI_PROP_MON::screen_list_init(
    UINT32 num_screens,
    sSCREEN_INFO *screens)
{
    m_screen_count = num_screens;

    if (m_screen_count > MAX_MON_COUNT)
    {
        mTERA_LOG_DBG("Unsupported number of screens: %u", m_screen_count);
        m_screen_count = 0;
    }
    else
    {
        for (UINT32 i = 0; i < m_screen_count; i++)
        {
            m_screen_list[i].width  = screens->width;
            m_screen_list[i].height = screens->height;
            m_screen_list[i].pos_x  = screens->pos_x;
            m_screen_list[i].pos_y  = screens->pos_y;
            screens++;
        }
    }
    mTERA_LOG_DBG("end");
    dump_dbg_info();
} // cPCOIP_AGENT_GUI_PROP_MON::screen_list_init()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::get_idx_monitor
//      Get the index of a monitor from the monitor list
//
//  Description:
//      This function returns the index of the Nth monitor from the monitor
//      list. The order of the monitors in the monitor list matches the order
//      of the monitors connected to the client. On a 4 port client a user can
//      connect monitors to ports 1 and 4. In this scenario, if the caller
//      wants the index of the 2nd monitor this function would return the value
//      3 (the index of monitor 4, where the valid indexes are 0, 1, 2 and 3).
//
//  Return:
//      TERA_SUCCESS when a valid monitor index is found
//      TERA_ERR_INVALID_STATE when the monitor index could not determined
//
UINT32 cPCOIP_AGENT_GUI_PROP_MON::get_idx_monitor(
    UINT32 mon_num,
    UINT32 *topo_idx)
{
    UINT32 connected_monitors = 0;

    // Initialize the monitor list if necessary
    if (m_mon_count == 0)
    {
        mon_list_init();
    }
    for (UINT32 i = 0; i < MAX_MON_COUNT; i++)
    {
        // if a monitor is connected display the monitor data
        if (is_client_mon_connected(i))
        {
            connected_monitors++;
            if (connected_monitors == mon_num)
            {
                mTERA_LOG_DBG("\tMonitor %d", i+1);
                mTERA_LOG_DBG("\t\tName:         %S", m_mon_list[i].name.c_str());
                mTERA_LOG_DBG("\t\tVendorID:     %S", m_mon_list[i].vid.c_str());
                mTERA_LOG_DBG("\t\tProductID:    %S", m_mon_list[i].pid.c_str());
                mTERA_LOG_DBG("\t\tSerialNumber: %S", m_mon_list[i].serial.c_str());
                mTERA_LOG_DBG("\t\tResolution:   %S", m_mon_list[i].resolution.c_str());
                *topo_idx = i;
                return TERA_SUCCESS;
            }
        }
    }

    mTERA_LOG_ERR("Unable to find monitor number %d", mon_num);
    return TERA_ERR_INVALID_STATE;
} // cPCOIP_AGENT_GUI_PROP_MON::get_idx_monitor()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::init_from_usr_cfg
//      Populates the screen information fields using configuration data
//
//  Description:
//      This function reads the configuration file and updates the
//      internal structures and the screen fields on the grid.
//
//      If only one screen is active and only one monitor is attached to the
//      client there is no need to read from the configuration file.
//
void cPCOIP_AGENT_GUI_PROP_MON::init_from_usr_cfg(void)
{
    UINT32                                 mon_count;
    UINT32                                 screen_count;
    wxString                               name;
    wxString                               vid;
    wxString                               pid;
    wxString                               serial;
    INT32                                  size_x, size_y;
    INT32                                  pos_x, pos_y;
    UINT32                                 monitor;
    UINT32                                 num_matches = 0;
    wxString                               label;
    wxString                               val;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;

    // if there is only one monitor and one screen we don't need valid
    // monitor configuration data
    if((m_mon_count == 1) && (m_screen_count == 1))
    {
        // Add this into the list of monitors
        m_screen_list[0].mon = 1;
        mTERA_LOG_DBG("one screen and monitor detected");
    }
    else
    {
        // Get number of valid monitors and screens in the user configuration
        mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
        wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);
        wxGetApp().gui().host_cfg().get_monitor_screen_counts(sess_link.mac,
                                                              &mon_count,
                                                              &screen_count);

        //
        // Proceed with reading the monitor information in the config file if
        // the configuration file monitor and screen counts match the number
        // of monitors reported by the FW and the number of active screens
        //
        if ((mon_count == m_mon_count) && (screen_count == m_screen_count))
        {
            for (UINT32 i = 0; i < screen_count; i++)
            {
                wxGetApp().gui().host_cfg().get_monitor_info(sess_link.mac,
                                                             i+1,
                                                             &vid,
                                                             &pid,
                                                             &serial,
                                                             &size_x,
                                                             &size_y,
                                                             &pos_x,
                                                             &pos_y,
                                                             &monitor);

                // TBD: possibly enhance the code by also verifying the position values
                // were populated from the config file. Currently the code does not
                // do this because the DEFAULT values for these fields are the same as
                // some valid values

                // Ignore any incomplete entries
                if ((vid == PCOIP_AGENT_CFG_MONX_VID_DEFAULT) ||
                    (pid == PCOIP_AGENT_CFG_MONX_PID_DEFAULT) ||
                    (serial == PCOIP_AGENT_CFG_MONX_SERIAL_DEFAULT) ||
                    (size_x == PCOIP_AGENT_CFG_MONX_SIZE_X_DEFAULT) ||
                    (size_y == PCOIP_AGENT_CFG_MONX_SIZE_Y_DEFAULT) ||
                    (monitor == PCOIP_AGENT_CFG_MONX_MON_DEFAULT))
                {
                    continue;
                }

                //
                // compare the monitor information read from the config file
                // with each monitor reported by the FW until a match is found
                //
                for (UINT32 k = 0; k < MAX_MON_COUNT; k++)
                {
                    //
                    // if the monitor reported by FW is found in the config
                    // file use the size, position and screen data found in the
                    // config file
                    //
                    if ((vid == m_mon_list[k].vid) &&
                        (pid == m_mon_list[k].pid) &&
                        (serial == m_mon_list[k].serial))
                    {
                        mTERA_LOG_DBG("screen %u data retrieved from config file", i);
                        mTERA_LOG_DBG("\tVendorID:     %S", vid.c_str());
                        mTERA_LOG_DBG("\tProductID:    %S", pid.c_str());
                        mTERA_LOG_DBG("\tSerialNumber: %S", serial.c_str());
                        mTERA_LOG_DBG("\tSize:         %ux%u", size_x, size_y);
                        mTERA_LOG_DBG("\tPosition:     (%u, %u)", pos_x, pos_y);
                        mTERA_LOG_DBG("\tMonitor:       %u", monitor);

                        // Add this into the list of monitors
                        m_screen_list[i].width = size_x;
                        m_screen_list[i].height = size_y;
                        m_screen_list[i].pos_x = pos_x;
                        m_screen_list[i].pos_y = pos_y;
                        m_screen_list[i].mon = monitor;

                        // increment the counter that keeps track of the number of matched monitors
                        num_matches++;
                    } // end if
                } // end for
            } // end for
        } // end if
        else
        {
            mTERA_LOG_DBG("cfg monitor or screen counts do not match counts detected");
        }

        // if there were problems reading the configuration data delete it
        if (num_matches != screen_count)
        {
            mTERA_LOG_ERR("Erasing user config data: matches %d, screen_count %d",
                          num_matches,
                          screen_count);
            //
            // delete the configuration file settings and reset the selected
            // monitor for each screen
            //
            delete_mon_usr_cfg();
            for (UINT32 i = 0; i < MAX_MON_COUNT; i++)
            {
                m_screen_list[i].mon = 255;
            }

            // clear flag indicating valid topology data has been sent to the client
            wxGetApp().gui().set_topo_sent_status(0);
        }
    } // end if
} // cPCOIP_AGENT_GUI_PROP_MON::init_from_usr_cfg()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate
//      Populate the property grid with the latest monitor information
//
//  Description:
//      This function populates the property grid with the latest monitor
//      information. The caller has the option of telling the function to
//      not read the config settings from the configuration file. This
//      is used when the configuration file settings are updated.
//
void cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate(int ignore_cfg_file)
{
    mTERA_LOG_DBG("start - ignore_cfg_file=%d", ignore_cfg_file);

    // First clear any existing entries
    m_prop_grid->Clear();

    // Init variables containing information on the screens and monitors
    mon_list_init();

    if(!ignore_cfg_file)
    {
        init_from_usr_cfg();
    }

    // Create the property grid entries
    prop_grid_entries_create();

    // Update the property grid with the values
    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        prop_grid_populate_screen_entry(i);
    }

    if (is_screen_cfg_position_ok())
    {
        // Populate the monitor information
        prop_grid_populate_monitor_info();

        // Setup and show the help tooltips
        if (m_mon_count)
        {
            help_btn_setup(0);
        }
        else
        {
            help_btn_setup(1);
        }
    }
    else
    {
        help_btn_setup(1);
    }

    //
    // If there is only one monitor and one screen write the config data. This
    // must be done to deal with the situation where the agent starts on a
    // client with 1 monitor and the config file does not have any information
    // about the monitor. After the config file is written the monitor info
    // is sent to the FW and the user has the option of enabling the feature.
    //
    if((m_mon_count == 1) && (m_screen_count == 1))
    {
        save_config();
    }

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::prop_grid_entries_create
//      Create the entries for the two property grids
//
//  Description:
//      This function creates the entries for the two property grids.
//
void cPCOIP_AGENT_GUI_PROP_MON::prop_grid_entries_create(void)
{
    UINT32 i;
    UINT32 j;
    wxString label;
    wxString val;
    wxArrayString arrMonitor;

    // create an array of the monitor choices
    arrMonitor.Add(_("<select monitor>"));
    for (j = 0; j < m_mon_count; j++)
    {
        label.Printf(_T("Monitor %u"), j+1);
        arrMonitor.Add(label);
    }

    m_prop_grid->Append(new wxPropertyCategory(_("Select a unique monitor for each X Window screen")));
    for (i = 0; i < m_screen_count; i++)
    {
        val.Printf(_T("Screen %u - %ux%u at (%u, %u)"),
                      i+1, m_screen_list[i].width, m_screen_list[i].height,
                      m_screen_list[i].pos_x, m_screen_list[i].pos_y);
        label.Printf(_T("screen%u_mon"), i);
        m_prop_grid->Append(new wxEnumProperty(val, label, arrMonitor));

        mTERA_LOG_DBG("m_screen_list[%u] - %ux%u at (%u, %u)",
                      i, m_screen_list[i].width, m_screen_list[i].height,
                      m_screen_list[i].pos_x, m_screen_list[i].pos_y);
    }

    // populate the column headings for the client monitor information grid
    m_client_mon_grid->SetColLabelValue(NAME_COLUMN,          _T("Name"));
    m_client_mon_grid->SetColLabelValue(VENDOR_ID_COLUMN,     _T("Vendor\nID"));
    m_client_mon_grid->SetColLabelValue(PRODUCT_ID_COLUMN,    _T("Product\nID"));
    m_client_mon_grid->SetColLabelValue(SERIAL_NUMBER_COLUMN, _T("Serial\nNumber"));
    m_client_mon_grid->SetColLabelValue(RESOLUTION_COLUMN,    _T("Size"));

    // Delete unused rows. This happens when the client has less than 4
    // monitors attached.
    if (m_mon_count < m_mon_row_entries)
    {
        mTERA_LOG_DBG("Deleting %u rows", m_mon_row_entries - m_mon_count);
        m_client_mon_grid->DeleteRows(m_mon_count, m_mon_row_entries - m_mon_count);
        m_mon_row_entries = m_mon_count;
    }

    // Add rows if needed. This happens when additional monitors are connected
    // to the client while a session is active. It also happens when a
    // session starts with a client with more monitors than the previously
    // connected client.
    if (m_mon_row_entries < m_mon_count)
    {
        mTERA_LOG_DBG("Adding %u rows", m_mon_count - m_mon_row_entries);
        m_client_mon_grid->AppendRows(m_mon_count - m_mon_row_entries);
        m_mon_row_entries = m_mon_count;
    }

    // populate the first entry of each row
    for (i = 0; i < m_mon_count; i++)
    {
        label.Printf(_T("Monitor %u"), i+1);
        m_client_mon_grid->SetRowLabelValue(i, label);
    }

    // Fit first column to width of longest label
    m_prop_grid->FitColumns();
} // cPCOIP_AGENT_GUI_PROP_MON::prop_grid_entries_create()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate_screen_entry
//      Update the specified screen entry on property grid
//
//  Description:
//      This function updates the property grid for the specified screen
//      with the latest monitor information. If no monitor is assigned to the
//      screen the entry is highlighted.
//
void cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate_screen_entry(
    UINT32 screen_index
    )
{
    wxString label;
    wxString val;

    mTERA_LOG_DBG("screen %d, monitor %d", screen_index, m_screen_list[screen_index].mon);

    label.Printf(_T("screen%u_mon"), screen_index);

    if ((m_screen_list[screen_index].mon > 0) &&
        (m_screen_list[screen_index].mon <= MAX_MON_COUNT))
    {
        val.Printf(_T("Monitor %d"), m_screen_list[screen_index].mon);
        m_prop_grid->SetPropertyValue(label, val);
        m_prop_grid->SetPropertyColourToDefault(label);
    }
    else
    {
        m_prop_grid->SetPropertyValue(label, _("<select monitor>"));
        m_prop_grid->SetPropertyTextColour(label, wxColour(255, 0, 0));
    }
} // cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate_screen_entry()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate_monitor_info
//      Populates the Monitor information fields on the grid
//
//  Description:
//      This function populates the monitor information fields on the grid
//      with the information stored in the m_mon_list structure.
//
void cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate_monitor_info(void)
{
    UINT32 mon_entry_idx = 0;
    mTERA_LOG_DBG("firmware reported %u monitors", m_mon_count);

    for (UINT32 i = 0; i < MAX_MON_COUNT; i++)
    {
        // if a monitor is connected update the grid with client monitor
        // information
        if (is_client_mon_connected(i))
        {
            if (mon_entry_idx < m_mon_row_entries)
            {
                m_client_mon_grid->SetCellValue(mon_entry_idx, NAME_COLUMN, m_mon_list[i].name);
                m_client_mon_grid->SetCellValue(mon_entry_idx, VENDOR_ID_COLUMN, m_mon_list[i].vid);
                m_client_mon_grid->SetCellValue(mon_entry_idx, PRODUCT_ID_COLUMN, m_mon_list[i].pid);
                m_client_mon_grid->SetCellValue(mon_entry_idx, SERIAL_NUMBER_COLUMN, m_mon_list[i].serial);
                m_client_mon_grid->SetCellValue(mon_entry_idx, RESOLUTION_COLUMN, m_mon_list[i].resolution);
                mon_entry_idx++;
            }
        }
    }

    // the grid now contains data so resize it and disable cell editing
    m_client_mon_grid->AutoSize();
    m_client_mon_grid->EnableEditing(false);

} // cPCOIP_AGENT_GUI_PROP_MON::prop_grid_populate_monitor_info()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::help_btn_setup
//      Set up and show the tooltip help button
//
//  Description:
//      This function sets up and shows the tooltip help button. The button
//      is formatted to look like plain text. When the user hovers or clicks
//      the text, a tooltip will appear. The tooltip provides possible
//      causes and solutions to why certain features are unavailable.
//
void cPCOIP_AGENT_GUI_PROP_MON::help_btn_setup(int display_help)
{
    if (!display_help)
    {
        // Hide the help text
        m_help_link->Show(false);
    }
    else
    {
        // Show the help text and set it to something generic first
        m_help_link->SetToolTip(IDS_STR_HELP_ZERO_MON);

        // ...then determine the best help text if it exists
        if (!wxGetApp().gui().pcoip_link().is_ready())
        {
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_LINK);
        }
        else if (m_cache_topo_read_err)
        {
            m_help_link->SetToolTip(IDS_STR_HELP_MON_CACHE_TOPO_ERR);
        }
        else if (!is_screen_cfg_position_ok())
        {
            m_help_link->SetToolTip(IDS_STR_HELP_MON_SCREEN_POS_ERR);
        }
        m_help_link->Show(true);
    }
} // cPCOIP_AGENT_GUI_PROP_MON::help_btn_setup()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::cPCOIP_AGENT_GUI_PROP_MON
///     Constructor
//
//  Description:
///     This is the PCoIP Host Software Network property page class constructor.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_MON::cPCOIP_AGENT_GUI_PROP_MON(
    wxWindow *parent
    ) :
    cPCOIP_AGENT_GUI_PROP_MON_BASE(parent)
{
    mTERA_LOG_DBG("start");

    m_screen_count = 0;
    m_cache_topo_read_err = 0;
    mon_list_clear();
    screen_list_clear();
    m_mon_row_entries = MAX_MON_COUNT;  // The GUI initializes the size of the grid

    m_apply_btn->SetToolTip(_("Apply monitor configuration settings"));

    // increase the height of the column label row, this is only done once
    m_client_mon_grid->SetColLabelSize(m_client_mon_grid->GetColLabelSize() + 10);

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_MON::cPCOIP_AGENT_GUI_PROP_MON()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::~cPCOIP_AGENT_GUI_PROP_MON
///     Destructor
//
//  Description:
///     This is the PCoIP Host Software Monitor property page class
///     destructor. This function cleans up any loose ends before being
///     destroyed.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_MON::~cPCOIP_AGENT_GUI_PROP_MON(void)
{
    mTERA_LOG_DBG("start");

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_MON::~cPCOIP_AGENT_GUI_PROP_MON()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::num_client_monitors
//      Returns the number of monitors reported by the client
//
//  Description:
//      This function returns the number of monitors reported by the client
//
// TBD - it looks like this function is not used, it should be deleted if it
//       isn't needed
int cPCOIP_AGENT_GUI_PROP_MON::num_client_monitors(void)
{
    return m_mon_count;
} // cPCOIP_AGENT_GUI_PROP_MON::num_client_monitors()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::is_client_mon_connected
//      Returns non-zero value if a monitor is connected to the specified index
//
//  Description:
//      This function looks at an entry in the m_mon_list array and returns
//      a non-zero value if a monitor is connected.
//
int cPCOIP_AGENT_GUI_PROP_MON::is_client_mon_connected(int index)
{
    if (((m_mon_list[index].name.length())) ||
        ((m_mon_list[index].vid.length())) ||
        ((m_mon_list[index].pid.length())) ||
        ((m_mon_list[index].serial.length())))
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // cPCOIP_AGENT_GUI_PROP_MON::is_client_mon_connected()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::is_screen_cfg_ok
//      Returns non-zero value if the screen configuration data is okay
//
//  Description:
//      This function returns a non-zero value if the screen configuration
//      settings are valid.
//
int cPCOIP_AGENT_GUI_PROP_MON::is_screen_cfg_ok(void)
{
    mon_list_init();
    init_from_usr_cfg();

    if (m_screen_count == 0)
    {
        mTERA_LOG_DBG("Screen config data not valid: zero screens");
        return 0;
    }

    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        if (m_screen_list[i].mon == 255)
        {
            mTERA_LOG_DBG("Screen config data not valid: screen %d invalid", i);
            return 0;
        }
    }

    mTERA_LOG_DBG("Screen config OK: num screens %d", m_screen_count);
    return 1;
} // cPCOIP_AGENT_GUI_PROP_MON::is_screen_cfg_ok()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_MON::is_screen_cfg_position_ok
//      Returns non-zero value if the screen configuration position data is okay
//
//  Description:
//      This function returns a non-zero value if the screen configuration
//      position settings are valid. If multiple monitors are active only one
//      screen should be located at position (0,0). This can happen if the
//      system is configured to have multiple active X Screens without Twinview
//      or Xinerama active.
//
int cPCOIP_AGENT_GUI_PROP_MON::is_screen_cfg_position_ok(void)
{
    int num_primary_monitors = 0;

    for (UINT32 i = 0; i < m_screen_count; i++)
    {
        if ((m_screen_list[i].pos_x == 0) &&
            (m_screen_list[i].pos_y == 0))
        {
            num_primary_monitors++;
        }
    }

    if (num_primary_monitors != 1)
    {
        mTERA_LOG_ERR("%d screens are located at (0,0)!", num_primary_monitors);
        return 0;
    }
    else
    {
        return 1;
    }
} // cPCOIP_AGENT_GUI_PROP_MON::is_screen_cfg_position_ok()

