////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Link Class
///
///     This file defines the PCoIP link class and related structures and
///     callbacks.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include "tera_pcoip_hcc_apdu.h"
#include "tera_pcoip_hdc_apdu.h"
#include "tera_dev.h"
#include "tera_disp.h"

// ---- Constants and Types -------------------------------------------------

#define TERA_PCOIP_LINK_MAX_NUM_DISP  (4)
#define TERA_DISP_IDEN_VID_LEN         4
#define TERA_DISP_IDEN_PID_LEN         5
#define TERA_DISP_IDEN_SERIAL_LEN      14
#define TERA_DISP_IDEN_NAME_LEN        14

typedef enum
{
    TERA_PCOIP_LINK_DISP_ROTATION_0   = 0,
    TERA_PCOIP_LINK_DISP_ROTATION_90  = 1,
    TERA_PCOIP_LINK_DISP_ROTATION_180 = 2,
    TERA_PCOIP_LINK_DISP_ROTATION_270 = 3,
} eTERA_PCOIP_LINK_DISP_ROTATION;

typedef struct sTERA_PCOIP_LINK_DISP_STRUCT
{
    INT32                          x_origin;
    INT32                          y_origin;
    UINT16                         width;
    UINT16                         height;
    eTERA_PCOIP_LINK_DISP_ROTATION rotation;
} sTERA_PCOIP_LINK_DISP;

typedef struct sTERA_PCOIP_LINK_DISP2_STRUCT
{
    INT32                          x_origin;
    INT32                          y_origin;
    UINT16                         width;
    UINT16                         height;
    eTERA_PCOIP_LINK_DISP_ROTATION rotation;
    UINT8                          refresh_rate;
} sTERA_PCOIP_LINK_DISP2;

typedef struct sTERA_DISP_IDENTITY_STRUCT
{
    char vid[TERA_DISP_IDEN_VID_LEN];
    char pid[TERA_DISP_IDEN_PID_LEN];
    char serial[TERA_DISP_IDEN_SERIAL_LEN];
    char name[TERA_DISP_IDEN_NAME_LEN];
} sTERA_DISP_IDENTITY;

typedef struct sTERA_PCOIP_LINK_DISP_TOPO_STRUCT
{
    UINT32                num_disp;
    sTERA_PCOIP_LINK_DISP disp[TERA_PCOIP_LINK_MAX_NUM_DISP];
} sTERA_PCOIP_LINK_DISP_TOPO;

typedef struct sTERA_PCOIP_LINK_CACHE_DISP_TOPO_STRUCT
{
    UINT32                 num_disp;
    sTERA_PCOIP_LINK_DISP2 disp[TERA_PCOIP_LINK_MAX_NUM_DISP];
    UINT32                 primary;
    sTERA_DISP_IDENTITY    iden[TERA_PCOIP_LINK_MAX_NUM_DISP];
} sTERA_PCOIP_LINK_CACHE_DISP_TOPO;

typedef struct sTERA_PCOIP_LINK_IMG_USER_CFG_STRUCT
{
    UINT32      image_qual_pref;           // frame rate versus quality
    UINT32      build_to_lossless_enable;
} sTERA_PCOIP_LINK_IMG_USER_CFG;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_LINK_KMP_EVT_ALG_TYPE
///     KMP event processing algorithm types
//
//  Description:
///     This enumeration defines all KMP event processing algorithm types.
///
typedef enum
{
    TERA_PCOIP_LINK_KMP_TSTAMP_ALG = 0,    ///< Inject events based on the event timestamp
    TERA_PCOIP_LINK_KMP_MIN_DELAY_ALG = 1, ///< Inject events when they are received (minimal delay)
    TERA_PCOIP_LINK_KMP_MAX_ALG = 2,
} eTERA_PCOIP_LINK_KMP_EVT_ALG_TYPE;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_LINK_LCUR_EVT_ALG_TYPE
///     Local cursor event processing algorithm types
//
//  Description:
///     This enumeration defines all Local cursor event processing algorithm
///     types.
///
///     Note: The driver might inject relative cursor reports while the driver
///           is onfigured to inject absolute reports. This happens when the
///           flag field of the mouse report APDU received from the client
///           sets the relative report preferred bit.
///
typedef enum
{
    TERA_PCOIP_LINK_LCUR_ABS_ALG = 0,    ///< Inject cursor events as absolute reports
    TERA_PCOIP_LINK_LCUR_REL_ALG = 1,    ///< Inject cursor events as relative reports
    TERA_PCOIP_LINK_LCUR_MAX_ALG = 2,
} eTERA_PCOIP_LINK_LCUR_EVT_ALG_TYPE;

// --------------------------------------------------------------------------
//  eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE
///     Cursor shape bitmap types
//
//  Description:
///     This enumeration defines all KMP cursor shape bitmap types.
///
typedef enum
{
    TERA_PCOIP_LINK_KMP_CUR_BITMAP_ALPHA = 0,       ///< Alpha
    TERA_PCOIP_LINK_KMP_CUR_BITMAP_COLOR = 1,       ///< Color
    TERA_PCOIP_LINK_KMP_CUR_BITMAP_XOR = 2,         ///< XOR
    TERA_PCOIP_LINK_KMP_CUR_BITMAP_COMP_ALPHA = 3,  ///< Compressed Alpha
    TERA_PCOIP_LINK_KMP_CUR_BITMAP_COMP_COLOR = 4,  ///< Compressed Color
    TERA_PCOIP_LINK_KMP_CUR_BITMAP_COMP_XOR = 5,    ///< Compressed XOR
    TERA_PCOIP_LINK_KMP_MAX_CUR_BITMAP = 6,
} eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE;

// Determines the maximum number of cached bitmaps the host SW will store.
// Currently the zero client and soft clients (4.6) only support caching 16
// bitmaps, which is why this constant is 16.
#define MAX_CACHE_BMP                16

// Cached bitmap structure
typedef struct sTERA_PCOIP_LINK_BMP_DATA_STRUCT
{
    eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE    bitmap_type;
    UINT8                                   hot_x;
    UINT8                                   hot_y;
    UINT8                                   width;
    UINT8                                   height;
    UINT8                                   and_depth;
    UINT8                                   xor_depth;
    UINT8                                  *bmp;          ///< pointer to bitmap data
    UINT32                                  bmp_len;
} sTERA_PCOIP_BMP_DATA;

// Structure containing variables that manage the cache of bitmaps
typedef struct sTERA_PCOIP_LINK_BMP_CACHE_DATA_STRUCT
{
    INT32                       max_num_bmps;
    UINT8                       idx_last_stored_bmp;
    sTERA_PCOIP_BMP_DATA        bmp[MAX_CACHE_BMP];
} sTERA_PCOIP_BMP_CACHE_DATA;

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK
///     Link class
//
//  Description:
///     This is the link class, which is used by userspace applications to talk
///     to the blade and client devices via the device drivers.
///
class cTERA_PCOIP_LINK
{

public:

    // --- Public Functions -------------------------------------------------

    cTERA_PCOIP_LINK(void);

    ~cTERA_PCOIP_LINK(void);

    INT is_ready(void);

    void link_heartbeat(
        INT                                    *sess_link_change,
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK *sess_link);

    INT32 start(void);

    void stop(void);

    // HMI2

    INT32 local_info_get(
        UINT32 *dev_type);

    INT32 link_info_get_avail(void);

    INT32 link_info_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO *link_info);

    INT32 ipv4_cfg_get_avail(void);

    INT32 ipv4_cfg_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG *ip4_cfg);

    INT32 sess_link_get_avail(void);

    INT32 sess_link_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK *sess_link);

    INT32 sess_link_disc_avail(void);

    INT32 sess_link_disc_with_reason_avail(void);

    INT32 sess_link_disc(
        UINT32 reason);

    INT32 host_sess_state_get(
        eTERA_PCOIP_HMI_HOST_SESSION_STATE *host_sess_state);

    INT32 host_sess_state_set(
        eTERA_PCOIP_HMI_HOST_SESSION_STATE state);

    INT32 sess_tag_set_avail(void);

    INT32 sess_tag_version_1_key_scramble_feature_avail(void);

    INT32 sess_tag_set(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG *sess_tag);

    INT32 conn_port_get_avail(void);

    INT32 conn_port_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT *conn_port);

    INT32 sess_stat_get_avail(void);

    INT32 sess_stat_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT *sess_stat);

    INT32 sess_stat_reset_avail(void);

    INT32 sess_stat_reset(void);

    INT32 sess_tag_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG *sess_tag);

    INT32 is_sess_view_brokered(void);

    INT32 wol_cfg_get_avail(void);

    INT32 wol_cfg_get(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG *wol_cfg);

    INT32 wol_cfg_set_avail(void);

    INT32 wol_cfg_set(
        sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG *wol_cfg);

    INT32 cached_topo_get_avail(void);

    INT32 cached_topo_set_avail(void);

    INT32 unicode_set_avail(void);

    INT32 unicode_set(UINT32 enable);

    INT32 img_user_ext_cfg_get_avail(void);

    INT32 img_user_ext_cfg_set_avail(void);

    // KMP

    INT32 hdc_open(void);

    INT32 hdc_close(void);

    INT32 lcur_lkbd_dev_feat_stat_get(
        UINT32 *feat_stat);

    INT32 lcur_lkbd_last_feat_stat_get(
        UINT32 *feat_stat,
        UINT32 *feat_stat_cnt);

    INT32 lcur_pointer_location_set_if_new(
        UINT16 abs_x,
        UINT16 abs_y);

    INT32 lcur_pointer_location_set(
        UINT32 tstamp,
        UINT16 x_pos,
        UINT16 y_pos);

    INT32 lcur_pointer_visibility_set(
        UINT32 tstamp,
        UINT8  visible);

    INT32 lkbd_client_auto_repeat_avail(void);

    INT32 lkbd_repeat_rates_get(
        UINT16 *repeat_delay,
        UINT16 *repeat_interval);

    INT32 lkbd_repeat_rates_set(
        UINT16 repeat_delay,
        UINT16 repeat_interval);

    INT32 lkbd_kmp_alg_type_set(
        eTERA_PCOIP_LINK_KMP_EVT_ALG_TYPE alg_type);

    INT32 kmp_driver_cfg_set(
        eTERA_PCOIP_LINK_LCUR_EVT_ALG_TYPE lcur_mode);

    INT32 lcur_shape_bmp_send_avail(void);

#ifdef HOST_CUR_BMP_DBG
    INT32 print_alpha_bitmap(
        UINT8   width,
        UINT8   height,
        UINT8  *bitmap);

    INT32 print_color_bitmap(
        UINT8   width,
        UINT8   height,
        UINT8   and_depth,
        UINT8   xor_depth,
        UINT8  *bitmap);
#endif

#if 0
    void dump_cursor_mask(
        const char *mask_name,
        WORD        biBitCount,
        LONG        height,
        LONG        width,
        UINT8      *pbits);
#endif

    INT32 init_bitmap_cache(void);

    INT32 flush_bitmap_cache(void);

    INT32 is_bitmap_cached(
        UINT8  hot_x,
        UINT8  hot_y,
        UINT8  width,
        UINT8  height,
        UINT8  and_depth,
        UINT8  xor_depth,
        void  *bmp,
        UINT32 bmp_len,
        eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE bitmap_type,
        UINT8 *cache_idx);

    void add_bitmap_to_cache(
        UINT8  hot_x,
        UINT8  hot_y,
        UINT8  width,
        UINT8  height,
        UINT8  and_depth,
        UINT8  xor_depth,
        void  *bmp,
        UINT32 bmp_len,
        eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE bitmap_type);

    INT32 lcur_shape_bmp_send(
        UINT8  hot_x,
        UINT8  hot_y,
        UINT8  width,
        UINT8  height,
        UINT8  and_depth,
        UINT8  xor_depth,
        void  *bmp,
        UINT32 bmp_len,
        eTERA_PCOIP_LINK_KMP_CUR_BITMAP_TYPE bitmap_type);

    INT32 lcur_change_shape_send(
        UINT8  index);

    INT32 lcur_shape_size_valid(
        UINT16 width,
        UINT16 height,
        INT    color_mask_present);

    INT32 lcur_spd_avail(void);

    INT32 lcur_spd_get(
        UINT32 *lcur_spd);

    INT32 lcur_spd_set(
        UINT32 lcur_spd);

    INT32 lcur_lkbd_avail(void);

    INT32 lcur_lkbd_enable(void);

    INT32 lcur_lkbd_disable(void);

    INT32 lcur_idle_hide_avail(void);

    INT32 lcur_idle_hide_set(
        UINT32 time);

    INT32 is_client_sc(void);

    INT32 lcur_disable_avail(void);

//    INT32 kbd_data_get(sTERA_PCOIP_HDC_KMP_KBD_DATA_EVT *kbd_data);

    // IMG

    INT32 disp_topology_get(
        sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO *disp_topo);

    INT32 async_disp_topology_get(
        sTERA_PCOIP_LINK_DISP_TOPO *disp_topo,
        UINT32                     *id);

    INT32 cache_disp_topology_get(
        sTERA_PCOIP_LINK_CACHE_DISP_TOPO *disp_topo);

    INT32 cache_disp_topology_set(
        sTERA_DISP_TOPO            *host_disp_topo,
        sTERA_DISP_EDID            *host_disp_edid,
        UINT32                      swap_portrait_en);

    INT32 disp_topology_send(
        sTERA_PCOIP_LINK_DISP_TOPO *disp_topo,
        INT32                       primary_width,
        INT32                       primary_height,
        INT                         twinview_active);

    INT32 img_user_cfg_req_get(
        sTERA_PCOIP_LINK_IMG_USER_CFG *user_cfg);

    INT32 img_user_cfg_set(
        UINT32      image_qual_pref);

    INT32 img_user_cfg_cache_set(
        UINT32      image_qual_pref);

protected:

    // --- Protected Functions ----------------------------------------------

    void hmi_avail_cmd_get(void);

    void hmi_avail_cmd_reset(void);

    void hmi_avail_get_var_get(void);

    void hmi_avail_set_var_get(void);

    void kmp_pri_ctxt_reset(void);

    void kmp_pri_ctxt_get(void);

    INT32 flip_bitmap(
        UINT8  width,
        UINT8  height,
        UINT8  xor_depth,
        UINT32 bmp_len,
        UINT8 *src,
        UINT8 *dest,
        INT32 *is_alpha);

    // --- Protected Variables ----------------------------------------------

    INT                               m_started;            ///< Started flag
    cTERA_DEV                         m_hcc_dev;            ///< PCoIP HCC device interface
    cTERA_DEV                         m_hdc_dev;            ///< PCoIP HDC device interface
    sTERA_PCOIP_HCC_APDU_HMI_MAP      m_hmi_avail_cmd;      ///< Available HMI commands
    sTERA_PCOIP_HCC_APDU_HMI_MAP      m_hmi_avail_get_var;  ///< Available HMI variables (R)
    sTERA_PCOIP_HCC_APDU_HMI_MAP      m_hmi_avail_set_var;  ///< Available HMI variables (W)
    sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT m_kmp_pri_ctxt;       ///< KMP PRI context
    INT                               m_kmp_pri_ctxt_valid; ///< Flag used to determine whether the context data is valid
    sTERA_PCOIP_BMP_CACHE_DATA        m_kmp_bmp_cache;      ///< Cached bitmap data

}; // cTERA_PCOIP_LINK


// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

