////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Link Class Environment Feature Implementation
///
///     This file implements link class environment feature. All environment
///     feature related operations are dealt with in this file. This feature
///     offers:
///         - sending host monitor information
///         - getting/setting the blade's and client's IPv4 network config
///         - getting/setting OSD mouse speed
///         - getting session information
///         - password authentication against the blade
///         - getting/setting Wake-on-LAN settings
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_types.h"
#include "tera_log.h"
#include "tera_errors.h"
#include "tera_mem.h"
#include "tera_pcoip_link.h"
#include "tera_pcoip_hcc_ioctl.h"
#include "tera_dev.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// HMI commands string table
static const char *hmi_cmd_str_tbl[TERA_PCOIP_HCC_APDU_HMI_CMD_MAX] = {
    "HMI_CMD_AVAIL_CMD",
    "HMI_CMD_VAR_GET",
    "HMI_CMD_VAR_SET",
};

// HMI variables string table
static const char *hmi_var_str_tbl[TERA_PCOIP_HCC_APDU_HMI_VAR_MAX] = {
    "HMI_VAR_AVAIL_VAR",
    "HMI_VAR_LINK_INFO",
    "HMI_VAR_IPV4_CFG",
    "HMI_VAR_SESS_LINK",
    "HMI_VAR_SESS_TAG",
    "HMI_VAR_CONN_PORT",
    "HMI_VAR_SESS_STAT",
    "HMI_VAR_DEPRECATED",
    "HMI_VAR_WOL_CFG",
    "HMI_VAR_SESS_DISC",
    "HMI_VAR_SESS_TAG2",
    "HMI_VAR_CACHED_TOPO",
    "HMI_VAR_UNICODE",
    "HMI_VAR_IMG_USER_EXT_CFG",
};

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::hmi_avail_cmd_get
//      Get the set of available environment feature commands
//
//  Description:
//      This function gets the set of available environment feature commands
//      from the device via an ioctl.
//
void cTERA_PCOIP_LINK::hmi_avail_cmd_get(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_MAP hmi_avail_cmd;

    if (m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_AVAIL_CMD,
                              &hmi_avail_cmd,
                              sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP),
                              1000) != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to get available HMI commands");

        // Clear the current available commands and environment variables
        hmi_avail_cmd_reset();
    }
    else
    {
        mTERA_LOG_DBG("***************************************");

        mTERA_LOG_DBG("hmi_avail_cmd.map[0] '0x%08X'", hmi_avail_cmd.map[0]);

        for (UINT32 i = 0; i < TERA_PCOIP_HCC_APDU_HMI_CMD_MAX; i++)
        {
            if (hmi_avail_cmd.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(i)] &
                mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(i))
            {
                mTERA_LOG_DBG("%s", hmi_cmd_str_tbl[i]);
            }
        }

        mTERA_LOG_DBG("***************************************");

        if (hmi_avail_cmd.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_GET))
        {
            // Get the list of available HMI variables for reading
            hmi_avail_get_var_get();
        }

        if (hmi_avail_cmd.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_CMD_VAR_SET))
        {
            // Get the list of available HMI variables for writing
            hmi_avail_set_var_get();
        }

        // Store the set of available commands
        mTERA_MEM_COPY(&m_hmi_avail_cmd, &hmi_avail_cmd,
                       sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));
    } // end if
} // cTERA_PCOIP_LINK::hmi_avail_cmd_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::hmi_avail_cmd_reset
//      Reset the list of available HMI commands and variables
//
//  Description:
//      This function resets the list of available HMI commands and variables.
//
void cTERA_PCOIP_LINK::hmi_avail_cmd_reset(void)
{
    // Reset the list of available HMI commands
    mTERA_MEM_ZERO(&m_hmi_avail_cmd, sizeof(m_hmi_avail_cmd));

    // Reset the list of available HMI variables
    mTERA_MEM_ZERO(&m_hmi_avail_get_var, sizeof(m_hmi_avail_get_var));
    mTERA_MEM_ZERO(&m_hmi_avail_set_var, sizeof(m_hmi_avail_set_var));
} // cTERA_PCOIP_LINK::hmi_avail_cmd_reset()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::hmi_avail_get_var_get
//      Get the list of available HMI variables for reading
//
//  Description:
//      This function gets the list of available HMI variables for reading
//      from the device via an ioctl.
//
void cTERA_PCOIP_LINK::hmi_avail_get_var_get(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    mTERA_MEM_ZERO(&buf, sizeof(buf));

    // Fill out the ioctl buffer
    buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR;

    // Send a synch ioctl to the device
    if (m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                              &buf, sizeof(buf),
                              1000) != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET ioctl failed");

        // Clear the list of available HMI variables
        mTERA_MEM_ZERO(&m_hmi_avail_get_var, sizeof(m_hmi_avail_get_var));
    }
    else
    {
        mTERA_LOG_DBG("***************************************");

        mTERA_LOG_DBG("buf.data.avail_var.map[0] '0x%08X'",
                      buf.data.avail_var.map[0]);

        for (UINT32 i = 0; i < TERA_PCOIP_HCC_APDU_HMI_VAR_MAX; i++)
        {
            if (buf.data.avail_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(i)] &
                mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(i))
            {
                mTERA_LOG_DBG("%s", hmi_var_str_tbl[i]);
            }
        }

        mTERA_LOG_DBG("***************************************");

        //TODO If session link is available, get the session link

        // Store the new list of available HMI variables
        mTERA_MEM_COPY(&m_hmi_avail_get_var, &buf.data.avail_var,
                       sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));
    }
} // cTERA_PCOIP_LINK::hmi_avail_get_var_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::hmi_avail_set_var_get
//      Get the list of available HMI variables for writing
//
//  Description:
//      This function gets the list of available HMI variables for writing
//      from the device via an ioctl.
//
void cTERA_PCOIP_LINK::hmi_avail_set_var_get(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    mTERA_MEM_ZERO(&buf, sizeof(buf));

    // Fill out the ioctl input buffer
    buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_AVAIL_VAR;

    // Send a synch ioctl to the device
    if (m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
                              &buf, sizeof(buf),
                              1000) != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("TERA_PCOIP_IOCTL_ENV_VAR_SET ioctl failed");

        // Clear the available environment variables
        mTERA_MEM_ZERO(&m_hmi_avail_set_var, sizeof(m_hmi_avail_set_var));
    }
    else
    {
        mTERA_LOG_DBG("***************************************");

        mTERA_LOG_DBG("buf.data.avail_var.map[0] '0x%08X'",
                      buf.data.avail_var.map[0]);

        for (UINT32 i = 0; i < TERA_PCOIP_HCC_APDU_HMI_VAR_MAX; i++)
        {
            if (buf.data.avail_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(i)] &
                mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(i))
            {
                mTERA_LOG_DBG("%s", hmi_var_str_tbl[i]);
            }
        }

        mTERA_LOG_DBG("***************************************");

        // Store the new set of available environment variables
        mTERA_MEM_COPY(&m_hmi_avail_set_var, &buf.data.avail_var,
                       sizeof(sTERA_PCOIP_HCC_APDU_HMI_MAP));
    }
} // cTERA_PCOIP_LINK::hmi_avail_set_var_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::link_info_get_avail
///     Check if getting of host's network link information is available
//
//  Description:
///     This function lets the caller know if getting of the host's network
///     link information is avalable. This function should be called prior to
///     calling the function to get the network link information.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::link_info_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::link_info_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::link_info_get
///     Get the host's network link information
//
//  Description:
///     This function gets the host's network link information from the device
///     via an ioctl. If successful, the network link information is copied to
///     the user's buffer.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::link_info_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO *link_info
    )
{
    INT32                       ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    mTERA_LOG_DBG("start");

    if (link_info == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else if (link_info_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_GET ioctl failed (ret=%d)", ret);
        }
        else
        {
            // Copy the blade's IPv4 config to the user buffer
            mTERA_MEM_COPY(link_info, &buf.data.link_info,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_LINK_INFO));
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::link_info_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::ipv4_cfg_get_avail
///     Check if setting of blade's IPv4 network configuration is available
//
//  Description:
///     This function lets the caller know if setting of the blade's IPv4
///     network configuration is avalable. This function should be called
///     prior to calling the function to set the network configuration.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::ipv4_cfg_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::ipv4_cfg_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::ipv4_cfg_get
///     Get the blade's IPv4 network configuration
//
//  Description:
///     This function gets the blade's IPv4 network configuration from the
///     device via an ioctl. If successful, the network configuration is copied
///     to the user's buffer.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::ipv4_cfg_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG *ipv4_cfg
    )
{
    INT32                       ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    mTERA_LOG_DBG("start");

    if (ipv4_cfg == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (ipv4_cfg_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET ioctl failed %d",
                          ret);
        }
        else
        {
            // Copy the blade's IPv4 config to the user buffer
            mTERA_MEM_COPY(ipv4_cfg, &buf.data.ipv4_cfg,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_IPV4_CFG));
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::ipv4_cfg_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_link_get_avail
///     Check if getting the session link information is available
//
//  Description:
///     This function lets the caller know if getting the session link
///     information is avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_link_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_link_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_link_get
///     Get the session link information
//
//  Description:
///     This function gets the session link information from the device via an
///     ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::sess_link_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK *sess_link
    )
{
    INT32                       ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

//    mTERA_LOG_DBG("start");

    if (sess_link == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (sess_link_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET ioctl failed %d",
                          ret);
        }
        else
        {
            // Copy the session info to the user buffer
            mTERA_MEM_COPY(sess_link, &buf.data.sess_link,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK));
        }
    }

//    mTERA_LOG_DBG("end - MAC=%s", sess_link->mac);

    return ret;
} // cTERA_PCOIP_LINK::sess_link_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_link_disc_avail
///     Check if disconnecting the session link is available
//
//  Description:
///     This function lets the caller know if disconnecting the session link
///     is avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_link_disc_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_link_disc_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_link_disc_with_reason_avail
///     Check if disconnecting the session link with a reason is available
//
//  Description:
///     This function lets the caller know if disconnecting the session link
///     and supplying a reason is avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_link_disc_with_reason_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_link_disc_with_reason_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_link_disc
///     Disconnect the session link
//
//  Description:
///     This function disconnects the session link, if there is an active
///     session link.
///
///	    The function sets the SESS_DISC variable if the FW supports this
///     variable. This variable allows the user to supply a reason for
///     disconnecting. If the FW does not support setting the SESS_DISC
///     variable this function will try to set the SESS_LINK variable if
///     FW supports setting it.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::sess_link_disc(
	UINT32 reason
	)
{
    INT32                        ret=TERA_ERR_NOT_SUPPORTED;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;
	INT						     done=0;

	//
	// If the FW reports it supports the session disconnect command try to
	// disconnect the session using it.
	//
    if (sess_link_disc_with_reason_avail() == TERA_SUCCESS)
    {
        // Fill out the ioctl input buffer
	    mTERA_MEM_ZERO(&buf, sizeof(buf));
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_DISC;
		buf.data.sess_disc.reason_cause = reason;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
                                    &buf, sizeof(buf),
                                    1000);
		//
		// This may fail if the system is using an old version of the host SW
		// (3.1.11 or earlier) and a new version of the view agent dll, which
		// includes this file. In this situation this function should attempt
		// to disconnect the session by setting the session link state.
		//
		if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_SET ioctl failed %d", ret);
        }
        else
        {
            mTERA_LOG_DBG("disconnected session link by setting SESS_DISC");
			done = 1;
        }
    }

	//
	// Disconnect the session by setting the session link state if the FW does
	// not support the session disconect command or an error occurred while
	// sending the session disconnect command.
	//
    if (!done)
    {
		if (sess_link_disc_avail() != TERA_SUCCESS)
		{
			mTERA_LOG_ERR("not supported");

			ret = TERA_ERR_NOT_SUPPORTED;
		}
		else
		{
			// Fill out the ioctl input buffer
			mTERA_MEM_ZERO(&buf, sizeof(buf));
			buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK;

			// Send a synch ioctl to the device
			ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
										&buf, sizeof(buf),
										1000);
			if (ret != TERA_SUCCESS)
			{
				mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_SET ioctl failed %d", ret);
			}
			else
			{
				mTERA_LOG_DBG("disconnected session link by setting SESS_LINK");
			}
		}
    }

    return ret;
} // cTERA_PCOIP_LINK::sess_link_disc()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::host_sess_state_get
///     Get the host session state
//
//  Description:
///     This function retrieves the host session state from the driver.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::host_sess_state_get(
	eTERA_PCOIP_HMI_HOST_SESSION_STATE *host_sess_state
	)
{
    INT32                     ret;
    sTERA_PCOIP_HMI_LOCAL_VAR buf;

    // Fill out the ioctl input buffer
    mTERA_MEM_ZERO(&buf, sizeof(buf));
    buf.var = TERA_PCOIP_HMI_LOCAL_VAR_HOST_SESS_STATE;

    // Send a synch ioctl to the device
    ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_LOCAL_VAR_GET,
                                &buf, sizeof(buf),
                                1000);

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HCC_IOCTL_HMI_LOCAL_VAR_GET ioctl failed %d", ret);
    }
    else
    {
        *host_sess_state = buf.data.sess_state.host_sess_state;

//        mTERA_LOG_DBG("host session state=%d", *host_sess_state);
    }

    return ret;
} // cTERA_PCOIP_LINK::host_sess_state_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::host_sess_state_set
///     Set the host session state
//
//  Description:
///     This function sends a message to the driver informing it about the
///     current host session state. The function should be called when the
///     host session state changes which can occur because the user logged
///     off/on or the session becomes locked/unlocked.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::host_sess_state_set(
	eTERA_PCOIP_HMI_HOST_SESSION_STATE state
	)
{
    INT32                     ret;
    sTERA_PCOIP_HMI_LOCAL_VAR buf;

    if (state >= TERA_PCOIP_HMI_HOST_STATE_INVALID)
    {
        mTERA_LOG_ERR("Invalid host state=%d", state);
        return TERA_ERR_NOT_SUPPORTED;
    }

    // Fill out the ioctl input buffer
    mTERA_MEM_ZERO(&buf, sizeof(buf));
    buf.var = TERA_PCOIP_HMI_LOCAL_VAR_HOST_SESS_STATE;
    buf.data.sess_state.host_sess_state = state;

    // Send a synch ioctl to the device
    ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_LOCAL_VAR_SET,
                                &buf, sizeof(buf),
                                1000);

    if (ret != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("HCC_IOCTL_HMI_LOCAL_VAR_SET ioctl failed %d", ret);
    }
    else
    {
        mTERA_LOG_DBG("set host session state=%d", state);
    }

    return ret;
} // cTERA_PCOIP_LINK::host_sess_state_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_tag_set_avail
///     Check if setting of the session tag is available.
//
//  Description:
///     This function lets the caller know if setting of the session tag is
///     available. If available, tag of any type (whether GUID 64 byte, Random string,
///     version 1 key scramble, or MAC based) can be set for reserving a pri
///     on the host firmware.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_tag_set_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_tag_set_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_tag_version_1_key_scramble_feature_avail
///     Check if handling of 'SCS1' Version 1 tag for encryption key
//      scrambling is supported
//
//  Description:
///     This function lets the caller know if the host firmware has the
///     capabiilty of handling 'SCS1' Version 1 tag for encryption key
///     scrambling (security feature described in BUG 5754). Although
///     this feature is not supported by the firmware, any tag can be set.
///
///     This capability refers to parsing the following format and using
///     the scrambling keys for the encryption keys being transported on the SSIG
///     INVITE packet.
///
/// -----------------VERSION 1 TAG FORMAT----------------
/// |VERSION|  SERVER_KEY   |  CLIENT_KEY   |  SESSION_ID   |
/// ------------------------------------------------
/// |4 bytes|   32 bytes    |   32 bytes    |   8 bytes     | = 76 bytes
/// ------------------------------------------------
/// |4 chars|  base64 char  |  base64 char  |  base64 char  |
/// ------------------------------------------------
/// |'SCS1' | 43 + 1 chars  | 43 + 1 chars  | 11 + 1 chars  |
/// ------------------------------------------------
// |4 bytes|   44 bytes    |   44 bytes    |   12 bytes    | = 106 bytes
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_tag_version_1_key_scramble_feature_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG2)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG2)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_tag_version_1_key_scramble_feature_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_tag_set
///     Set the session tag
//
//  Description:
///     This function sets the session tag on the device via an ioctl. The
///     session tag is used by both endpoints to verify that a session
///     connection is valid.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::sess_tag_set(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG *sess_tag
    )
{
    INT32                       ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    if (sess_tag == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else if (sess_tag_set_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG;

        // Copy the session tag to the ioctl input buffer
        mTERA_MEM_COPY(&buf.data.sess_tag, sess_tag,
                       sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG));

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET ioctl failed %d",
                          ret);
        }
        else
        {
#if 1
            // don't display the tag for security purposes
            mTERA_LOG_DBG("session tag set: flags=0x%02X, length tag=%ld",
                          sess_tag->flags, strlen(sess_tag->tag));
#else
            mTERA_LOG_DBG("session tag: flags=0x%02X, tag=%s",
                          sess_tag->flags, sess_tag->tag);
#endif
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::sess_tag_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::conn_port_get_avail
///     Check if getting the connection port is available
//
//  Description:
///     This function lets the caller know if getting the connection port is
///     avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::conn_port_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::conn_port_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::conn_port_get
///     Get the connection port
//
//  Description:
///     This function gets the connection port from the device via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::conn_port_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT *conn_port
    )
{
    INT32                       ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    mTERA_LOG_DBG("start");

    if (conn_port == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (conn_port_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET ioctl failed %d",
                          ret);
        }
        else
        {
            // Copy the session info to the user buffer
            mTERA_MEM_COPY(conn_port, &buf.data.conn_port,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_CONN_PORT));
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::conn_port_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_stat_get_avail
///     Check if getting the session statistics is available
//
//  Description:
///     This function lets the caller know if getting the session statistics is
///     avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_stat_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_stat_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_stat_get
///     Get the session statistics
//
//  Description:
///     This function gets the session statistics from the device via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::sess_stat_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT *sess_stat
    )
{
    INT32                        ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

//    mTERA_LOG_DBG("start");

    if (sess_stat == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else if (sess_stat_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_GET ioctl failed (ret=%d)", ret);
        }
        else
        {
            // Copy the session info to the user buffer
            mTERA_MEM_COPY(sess_stat, &buf.data.sess_stat,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT));
        }
    }

//    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::sess_stat_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_stat_reset_avail
///     Check if resetting of the session statistics is available
//
//  Description:
///     This function lets the caller know if resetting of the session
///     statistics is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::sess_stat_reset_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::sess_stat_reset_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_stat_reset
///     Reset the session statistics
//
//  Description:
///     This function resets the session statistics on the device via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::sess_stat_reset(void)
{
    INT32                       ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    mTERA_LOG_DBG("start");

    if (sess_stat_reset_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");
        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET ioctl failed %d",
                          ret);
        }
    }

    mTERA_LOG_DBG("end");

    return ret;
} // cTERA_PCOIP_LINK::sess_stat_reset()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::sess_tag_get
///     Get a copy of the last session tag
//
//  Description:
///     This function gets a copy of the last session tag message. This message
///     is written to the host card by the View agent running on the host PC.
///     The agent looks at the session tag message to determine if the
///     connection was brokered.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::sess_tag_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG *sess_tag
    )
{
    INT32                        ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    if (sess_tag == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_GET ioctl failed (ret=%d)", ret);
        }
        else
        {
            // Copy the blade's session tag to the user buffer
            mTERA_MEM_COPY(sess_tag, &buf.data.sess_tag,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG));

            // Note: The driver dumps information about the tag. This is
            //       done intentionally because the View agent also reads
            //       the tag and I want to dump log messages when it does
            //       this.
#if 1
            // don't display the tag for security purposes
            mTERA_LOG_DBG("session tag get: flags=0x%02X, length tag=%ld",
                          sess_tag->flags, strlen(sess_tag->tag));
#endif
#if 0
            // don't display the tag for security purposes
            mTERA_LOG_DBG("session tag: flags=0x%02X, tag=%s",
                          sess_tag->flags, sess_tag->tag);
#endif
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::sess_tag_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::is_sess_view_brokered
///     Returns non-zero value if the session is view brokered, otherwise zero
//
//  Description:
///     This function reads the session tag to determine whether the current
///     session is brokered by a VMware View connection server.
///
/// \return
///     zero - session is NOT view brokered
///     non-zero - session is view brokered
///
INT32 cTERA_PCOIP_LINK::is_sess_view_brokered(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG sess_tag;

    if((sess_tag_get(&sess_tag) == TERA_SUCCESS) &&
       (sess_tag.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_TAG_FLAGS_RESERVE))
    {
        return 1;
    }
    else
    {
        return 0;
    }
} // cTERA_PCOIP_LINK::is_sess_view_brokered()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::wol_cfg_get_avail
///     Check if getting the Wake-on-LAN configuration is available
//
//  Description:
///     This function lets the caller know if getting the Wake-on-LAN (WOL)
///     configuration is avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::wol_cfg_get_avail(void)
{
    // If the session is View brokered the WOL feature will be disabled.
    if ((m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG)] &
         mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG)) &&
        !(is_sess_view_brokered()))
    {
        return TERA_SUCCESS;
    }
    else
    {
        return TERA_ERR_NOT_SUPPORTED;
    }
} // cTERA_PCOIP_LINK::wol_cfg_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::wol_cfg_get
///     Get the Wake-on-LAN configuration
//
//  Description:
///     This function gets the Wake-on-LAN (WOL) configuration from the device
///     via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::wol_cfg_get(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG *wol_cfg
    )
{
    INT32                        ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    if (wol_cfg == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else if (wol_cfg_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG;

        // Send a synch ioctl to the device. Note this operation can take
		// more than a second to complete on long latency connections (>120 ms).
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_GET,
                                    &buf, sizeof(buf),
                                    2000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_GET ioctl failed (ret=%d)", ret);
        }
        else
        {
            // Copy the session info to the user buffer
            mTERA_MEM_COPY(wol_cfg, &buf.data.wol_cfg,
                           sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG));
            mTERA_LOG_DBG("flags=0x%02X, IP=%s, MAC=%s",
				          wol_cfg->flags,
						  wol_cfg->ip,
						  wol_cfg->mac);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::wol_cfg_get()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::wol_cfg_set_avail
///     Check if setting the Wake-on-LAN configuration is available
//
//  Description:
///     This function lets the caller know if setting the Wake-on-LAN (WOL)
///     configuration is avalable.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::wol_cfg_set_avail(void)
{
    // If the session is View brokered the WOL feature will be disabled.
    if ((m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG)] &
         mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG)) &&
        !(is_sess_view_brokered()))
    {
        return TERA_SUCCESS;
    }
    else
    {
        return TERA_ERR_NOT_SUPPORTED;
    }
} // cTERA_PCOIP_LINK::wol_cfg_set_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::wol_cfg_set
///     Set the Wake-on-LAN configuration
//
//  Description:
///     This function sets the Wake-on-LAN (WOL) configuration from the device
///     via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::wol_cfg_set(
    sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG *wol_cfg
    )
{
    INT32                        ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    if (wol_cfg == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");

        ret = TERA_ERR_INVALID_ARG;
    }
    else if (wol_cfg_get_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG;

        // Copy the user supplied WOL data to the ioctl input buffer
        mTERA_MEM_COPY(&buf.data.wol_cfg, wol_cfg,
                       sizeof(sTERA_PCOIP_HCC_APDU_HMI_VAR_WOL_CFG));

        // Send a synch ioctl to the device. Note this operation can take
		// more than a second to complete on long latency connections (>120 ms).
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
                                    &buf, sizeof(buf),
                                    2000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_SET ioctl failed (ret=%d)", ret);
        }
		else
		{
            mTERA_LOG_DBG("flags=0x%02X, IP=%s, MAC=%s",
				          wol_cfg->flags,
						  wol_cfg->ip,
						  wol_cfg->mac);
		}
    }

    return ret;
} // cTERA_PCOIP_LINK::wol_cfg_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::cached_topo_get_avail
///     Check if getting of the cached topology is available
//
//  Description:
///     This function lets the caller know if getting of the cached topology
///     is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::cached_topo_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_CACHED_TOPO)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_CACHED_TOPO)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::cached_topo_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::cached_topo_set_avail
///     Check if setting of the cached topology is available
//
//  Description:
///     This function lets the caller know if setting of the cached topology
///     is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::cached_topo_set_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_CACHED_TOPO)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_CACHED_TOPO)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::cached_topo_set_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::unicode_set_avail
///     Check if setting of the unicode enable is available
//
//  Description:
///     This function lets the caller know if setting of the unicode enable
///     is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::unicode_set_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::unicode_set_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::unicode_set
///     Set the unicode configuration
//
//  Description:
///     This function sets the Host FW unicode configuration via an ioctl.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_STATE,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_NOT_SUPPORTED,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::unicode_set(
    UINT32 enable
    )
{
    INT32                        ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR buf;

    if (unicode_set_avail() != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("not supported");

        ret = TERA_ERR_NOT_SUPPORTED;
    }
    else
    {
        mTERA_MEM_ZERO(&buf, sizeof(buf));

        // Fill out the ioctl input buffer
        buf.var = TERA_PCOIP_HCC_APDU_HMI_VAR_UNICODE;
        buf.data.unicode.enabled = enable;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_VAR_SET,
                                    &buf, sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_VAR_SET ioctl failed (ret=%d)", ret);
        }
		else
		{
            mTERA_LOG_DBG("enable=%d", enable);
		}
    }

    return ret;
} // cTERA_PCOIP_LINK::unicode_set()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::img_user_ext_cfg_get_avail
///     Check if getting of the imaging user ext config is available
//
//  Description:
///     This function lets the caller know if getting of the imaging user
///     extended configuration is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::img_user_ext_cfg_get_avail(void)
{
    return (m_hmi_avail_get_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_IMG_USER_EXT_CFG)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_IMG_USER_EXT_CFG)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::img_user_ext_cfg_get_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::img_user_ext_cfg_set_avail
///     Check if setting of the imaging user ext config is available
//
//  Description:
///     This function lets the caller know if setting of the imaging user
///     extended configuration is available.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NOT_SUPPORTED
///
INT32 cTERA_PCOIP_LINK::img_user_ext_cfg_set_avail(void)
{
    return (m_hmi_avail_set_var.map[mTERA_PCOIP_HCC_APDU_HMI_MAP_INDEX(TERA_PCOIP_HCC_APDU_HMI_VAR_IMG_USER_EXT_CFG)] &
            mTERA_PCOIP_HCC_APDU_HMI_MAP_BITMASK(TERA_PCOIP_HCC_APDU_HMI_VAR_IMG_USER_EXT_CFG)) ?
           TERA_SUCCESS : TERA_ERR_NOT_SUPPORTED;
} // cTERA_PCOIP_LINK::img_user_ext_cfg_set_avail()

// --------------------------------------------------------------------------
//  cTERA_PCOIP_LINK::local_info_get
///     Get the local info from the device driver
//
//  Description:
///     This function gets the local information maintained by the driver.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_INVALID_ARG,
///     TERA_ERR_TIMEOUT,
///     TERA_ERR_FAILURE
///
INT32 cTERA_PCOIP_LINK::local_info_get(
    UINT32 *dev_type
    )
{
    INT32                     ret;
    sTERA_PCOIP_HMI_LOCAL_VAR buf;

    if (dev_type == NULL)
    {
        mTERA_LOG_ERR("invalid arguments");
        ret = TERA_ERR_INVALID_ARG;
    }
    else
    {
        // Fill out the ioctl input buffer
        mTERA_MEM_ZERO(&buf, sizeof(buf));
        buf.var = TERA_PCOIP_HMI_LOCAL_VAR_LOCAL_INFO;

        // Send a synch ioctl to the device
        ret = m_hcc_dev.synch_ioctl(TERA_PCOIP_HCC_IOCTL_HMI_LOCAL_VAR_GET,
                                    &buf,
                                    sizeof(buf),
                                    1000);
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_ERR("HCC_IOCTL_HMI_LOCAL_VAR_GET ioctl failed (ret=%d)",
                          ret);
        }
        else
        {
            *dev_type = buf.data.info.dev_type;
//            mTERA_LOG_DBG("TERA processor device type=%u", *dev_type);
        }
    }

    return ret;
} // cTERA_PCOIP_LINK::local_info_get()
